{$C6000} {$X+} {$G+} {$D+}
program focus_pocus (input,output);
{A program to calculate the focal length and beam waist produced by a lens 
 in a gaussian beam}
{Refer to "Focal length calculation" APG March 1986}

const pi=3.14159265;
      gborder=50;
      gheight=780;{graphics screen size}
      gwidth=1000;

type styles=(dot,nodot);
     string=packed array [1..32] of char;
      
var
  d1,d1a,d3,d3a,w01,w03,w03a,f_lens,lambda  :real;
  x,x1,x2                                  :real;
  horiz_scale,vert_scale                   :real;
  ipi                                      :integer;
  l                                        :real;
  ipc                                      :char;
  flag,neg1,neg3                           :boolean;
  s1,s3                                    :styles;

procedure tab(x,y:integer);
begin;
 vdu(31,x,y);
end;

procedure dump;
begin;
 tab(71,30);writeln('Copying');
 oscli('GIMAGE EPSON');
 tab(71,30);writeln('       ');
end;

procedure graph_plot(s:integer; x,y:real);
begin;
 plot(s,gborder+trunc(gwidth*x),gborder+trunc(gheight*y));
end;

procedure plot_anchor(x,y :real);
begin;
 graph_plot(4,x,y);
end;

procedure plot_relative(style:styles; x,y :real);
var    s :integer;
begin;
 case style of
 dot:s:=17;
 nodot:s:=1;
 end;
 plot(s,trunc(gwidth*x),trunc(gheight*y));
end;

procedure plotline(style:styles; h_end,y_end :real; var h_beg :real;
                    var y_beg :real);
var     s:integer;

begin;
 case style of
 dot:s:=21;
 nodot:s:=5;
 end;
 plot_anchor(h_beg,y_beg);
 graph_plot(s,h_end,y_end);
 h_beg:=h_end;
 y_beg:=y_end;
end;

procedure draw_x_axis;
var n         :integer;

begin;
 plot_anchor(0.0,0.5); 
 plot_relative(nodot,1.0,0.0);
 for n:=0 to 10 do
  begin;
  plot_anchor(n*0.1,0.5);
  plot(1,0,-10);
  end;
end;

function omega(z,w0,lambda :real):real;
begin;
 omega:=w0*sqrt( 1 + sqr(lambda*z/(pi*sqr(w0))) );
end;

procedure drawgauss(d,w0,lambda,l,horiz_scale,vert_scale:real;s:styles);
var
    h_old1,h_old2,y_old1,y_old2      :real;
    z,w,h                            :real;
    n,nt                             :integer;

begin;
 h_old1:=l*horiz_scale;
 h_old2:=h_old1;
 y_old1:=0.5+vert_scale;
 y_old2:=0.5-vert_scale;

 nt:=trunc(200.0*abs(d)*horiz_scale);
 for n:=nt downto 0 do
  begin;
  z:=n*d/nt;
  h:=(l-d+z)*horiz_scale;
  w:=omega(z,w0,lambda);
  plotline(s,h,w*vert_scale+0.5,h_old1,y_old1);
  plotline(s,h,0.5-w*vert_scale,h_old2,y_old2);
  end;
 
end;

begin;
repeat
 mode(0);
 writeln('GAUSSIAN BEAM LENS CALCULATOR');
 writeln;
 write('Wavelength = ');
 readln(lambda);
 write('Focal length of lens = ');
 readln(f_lens);
 write('w01 = ');
 readln(w01);
 write('d1 = ');
 readln(d1);
 flag:=false;
 mode(0);
repeat
  tab(0,0);
  writeln('LENS OF ',f_lens:3:3,' FOCAL LENGTH IN GAUSSIAN BEAM');
  writeln;
  x1:=1 - d1/f_lens;
  x2:=pi*sqr(w01)/(lambda);
  x:=sqr(x1) + sqr(x2/f_lens);
  w03:=w01/sqrt(x);
  d3:=(sqr(x2)/f_lens - d1*x1)/x;

  if flag then writeln('w01=',w01:3:3,' w03=',w03a:3:3,' d1:=',d1a:3:3,
                        ' d3=',d3a:3:3)
  else writeln('Virtual foci are dotted');

  writeln('w01=',w01:3:3,' w03=',w03:3:3,' d1=',d1:3:3,' d3=',d3:3:3);
  draw_x_axis;
  neg1:=(d1<0);neg3:=(d3<0);
  if neg1 then s1:=dot else s1:=nodot;
  if neg3 then s3:=dot else s3:=nodot;
  if (neg1 and neg3) or (not(neg1) and not(neg3)) then
  {foci on different sides of lens}
  begin;
   horiz_scale:=abs(1/(d1+d3));
   if neg1 then l:=-d3 else l:=d1;
  end
  else
  {foci on same side of lens}
  begin;
   if abs(d1)>abs(d3) then horiz_scale:=abs(1/d1) else horiz_scale:=abs(1/d3);
   if neg1 then l:=0 else l:=1/horiz_scale;
  end;

  vert_scale:=0.5/omega(d1,w01,lambda);

  tab(trunc((horiz_scale*(l-d1))*62)+1,17);
  writeln('w01');

  tab(trunc((horiz_scale*(l+d3))*62)+1,17);
  writeln(' w03  ');

  {draw lines at waists and lens}
  plot_anchor(horiz_scale*l,0.0);
  plot_relative(nodot,0.0,1.0);
  plot_anchor(horiz_scale*(l-d1),0.5-w01*vert_scale);
  plot_relative(s1,0.0,2*w01*vert_scale);
  plot_anchor(horiz_scale*(l+d3),0.5-w03*vert_scale);
  plot_relative(s3,0.0,2*w03*vert_scale);

  drawgauss(d1,w01,lambda,l,horiz_scale,vert_scale,s1);
  drawgauss(-d3,w03,lambda,l,horiz_scale,vert_scale,s3); 

 repeat
   tab(71,2);
   writeln('Quit');
   tab(71,4);                                                                      writeln('Copy');
   tab(71,6);
   writeln('Run');
   tab(71,8);
   writeln('d1 value');
  repeat

   repeat
     ipi:=inkey(10);
   until ipi > 0;

    ipc:=chr(ipi);
  until ipc in ['C','c','D','d','Q','q','R','r'];

    if ipc in ['C','c'] then dump;

 until ipc in ['D','d','Q','q','R','r'];

   if ipc in ['D','d'] then 
    begin;
     flag:=true;
     d1a:=d1;d3a:=d3;w03a:=w03;
     tab (70,29);
     writeln ('d1 =');
     tab (70,30);
     readln(d1);
     tab(70,29);
     writeln('    ');
     tab(70,30);
     writeln('        ');
    end;
                                                                            
until ipc in ['Q','q','R','r'];

until ipc in ['Q','q'];

end.
 