program Mobius_transformation (input,output,datafile);
{A program to make error corrections for multiple reflections 
                                              on reflectometers}

{Note for this program to work properly it is required that the statement
  write(chr(13)) should not produce a line feed. Two small modifications
  to the file runc1.mac were needed before assembling.}

const  pi=3.141592654;

type complex=record
                re  :real;
                im  :real; 
     end;
     
     polar=record
                mag :real;
                ph  :real;
     end;

     somechar='A'..'z';

string=packed array[1..7] of char;

var
    d1,d2,d3                  :polar;{Actual magnitudes and phases}
    m1,m2,m3                  :polar;{Measured magnitudes and phases}
    z1, z2, z3                :complex;{Actual impedances}
    w1, w2, w3                :complex;{Real impedances}
    wa, wb, za, zb, p         :complex;{Temporary variables}
    au, bu, cu, du            :complex;{Unnormalised coeffs}
    a, b, c                   :complex;{Normalised coeffs}
    k, unity                  :complex;
    num, den                  :complex;
    ipc, ipc1, ipc2           :somechar;
    cset, ckit                :somechar;
    datafile                  :file of real;
    calkitfn, calsetfn        :string;{filenames}
    sname                     :string;
    nm                        :packed array[1..3] of char;
    flag                      :boolean;

procedure cursorlast;
{Cursor to last line of screen}
var n :integer;
begin
  for n:=1 to 25 do writeln
end;

procedure oscli(s:string);
begin end;

procedure getkey(var ipc:somechar);
begin;
 read(ipc);
end;

procedure inverse;
begin;
 write(chr(27),'i');
end;

procedure not_inverse;
begin;
 write(chr(27),'n');
end;

procedure clearline;
{Delete current line}
begin;
 write(chr(13));
 write('                                                                   ');
 write(chr(13));
end;

procedure mul (a,b :complex; var c :complex);
begin;
  c.re:=(a.re*b.re - a.im*b.im);
  c.im:=(a.re*b.im + a.im*b.re);
end;

procedure dvd (a,b:complex; var c :complex);
var denom  :real;
begin;
 denom:=sqr(b.re)+sqr(b.im);
 c.re:=(b.re*a.re + b.im*a.im)/denom;
 c.im:=(b.re*a.im - b.im*a.re)/denom;
end;

procedure add (a,b:complex; var c:complex);
begin;
 c.re:=a.re + b.re;
 c.im:=a.im + b.im;
end;

procedure diff (a,b:complex; var c:complex);
begin;
 c.re:=a.re - b.re;
 c.im:=a.im - b.im;
end;

procedure rect_polar(c:complex; var b:polar);
{Note phase angle in degrees}
var x, y   : boolean;
begin;
 b.mag:=sqrt(sqr(c.re) + sqr(c.im));
 b.ph:=180.0*arctan(abs(c.im/c.re))/pi;
  {which quadrant?}
x:=(c.re<0.0);y:=(c.im<0.0);
if x and y then b.ph:=180.0 + b.ph;
if x and not(y) then b.ph:=180.0 - b.ph;
if not(x) and y then b.ph:=360.0 - b.ph;
end;

procedure polar_rect(b :polar; var c:complex);
{Note phase angle in degrees}
var  phrad    :real;
begin;
 phrad:=b.ph*pi/180.0;
 c.re:=b.mag*cos(phrad);
 c.im:=b.mag*sin(phrad);
end;

procedure standards(a,b,c:complex; d1,d2,d3,m1,m2,m3:polar;ckit,cset
                             :somechar);
begin;
   writeln;
   writeln('STANDARDS  CalKit ',ckit,'                      CalSet ',cset);
   writeln;
   writeln('1 - ','Actual: mag=',d1.mag:8:3,' phase=',d1.ph:8:3,
                    '  Meas: mag=',m1.mag:8:3,' phase=',m1.ph:8:3);
   writeln('2 - ','            ',d2.mag:8:3,'       ',d2.ph:8:3,
                    '            ',m2.mag:8:3,'       ',m2.ph:8:3);
   writeln('3 - ','            ',d3.mag:8:3,'       ',d3.ph:8:3,
                    '            ',m3.mag:8:3,'       ',m3.ph:8:3); 
   writeln;
   writeln('a = ',a.re:5:3,' + j',a.im:5:3);
   writeln('b = ',b.re:5:3,' + j',b.im:5:3);
   writeln('c = ',c.re:5:5,' + j',c.im:5:3);
   writeln;
    
end;

procedure readcmplx(var b :polar; var c :complex);
begin;
 write('   Magnitude=');
 readln(b.mag);
 write('   Phase=');
 readln(b.ph);
 polar_rect(b,c);
end;
     
procedure zfind(a,b,c:complex);
var w, z :complex;
    m, d :polar;
    p,q,r,num,den     :complex; {temporary variables}

begin;
writeln;
writeln('Measured ');
readcmplx(m,w);
unity.re:=1.0;unity.im:=0.0;

{Denominator}
mul(c,w,p);
add(p,unity,den);

{Numerator}
mul(w,a,p);
add(p,b,num);
{}
dvd(num,den,z);
rect_polar(z,d);


writeln('Actual magnitude=',d.mag:3:3,' phase=',d.ph:3:3);
writeln;
end;

begin;{Main program}
 page;
  cursorlast;
  writeln('ERROR CORRECTION PROGRAM');
  writeln;
  ckit:='N';
  cset:='N';{Indicates unset}

 repeat {K option loop}
  repeat
   inverse;
   write('L-Load Calkit    D-Define Calkit ');
   if ckit<>'N' then write(' Q-Quit ');
   not_inverse;
  repeat
    getkey(ipc); 
  until (ipc in ['L','l','D','d']) or ((ckit='N') and (ipc in ['Q','q']));
   clearline;
                                                                              
   case ipc of
   'D','d': begin;
    ckit:='*';
    flag:=true;
    writeln;
    writeln(' Define three standards:');
    writeln('Definition 1');readcmplx(d1,z1);
    writeln('Definition 2');readcmplx(d2,z2);
    writeln('Definition 3');readcmplx(d3,z3);
    writeln;
    
   end;

   'L','l': begin;
     inverse;
     write('Quit  Load CalKit number 1-9 ? ');
     not_inverse;
    repeat
      getkey(ipc1);
    until ipc1 in ['1'..'9','Q','q'];
     clearline;
     flag:=false;
     if ipc1 in ['1'..'9'] then
     begin;
      inverse;
      write('Load CalKit ',ipc1,'  <Return> ');
      not_inverse;
      getkey(ipc2);  {A problem here - cannot get CR without inkey}
      clearline;
      if ipc2=chr(13) then
      begin;
       ckit:=ipc1;
       flag:=true;
       calkitfn:='calkit ';
       calkitfn[7]:=ckit;
       reset(datafile {,calkitfn} );
       read(datafile,d1.mag,d1.ph,z1.re,z1.im);
       read(datafile,d2.mag,d2.ph,z2.re,z2.im);
       read(datafile,d3.mag,d3.ph,z3.re,z3.im);
      end;
     end; 
    end;

    'Q','q': flag:=true;

    end; {Case}

   until flag=true;

 repeat {S option loop}
  repeat
   inverse;
   write('L-Load Calset  D-Define Calset ');
   if cset<>'N' then write(' Q-Quit ');
   not_inverse;
  repeat
    getkey(ipc);
  until (ipc in ['D','d','L','l']) or ((cset<>'N') and (ipc in ['Q','q']));
   clearline;

   case ipc of
   'D','d': begin; 
    flag:=true;
    cset:='*';
    writeln;
    writeln(' Enter measured standards');
    writeln('Measured 1');readcmplx(m1,w1);
    writeln('Measured 2');readcmplx(m2,w2);
    writeln('Measured 3');readcmplx(m3,w3);
   end;

   'L','l': begin;
    inverse;
    write('Quit  Load CalSet number 1-9 ? ');
    not_inverse;
    flag:=false;
   repeat
     getkey(ipc1);
   until ipc1 in ['Q','q','1'..'9'];
    clearline;
    if ipc1 in ['1'..'9'] then
    begin;
     inverse;
     write('Load CalSet ',ipc1,'  <Return> ');
     not_inverse;
     getkey(ipc2);
     clearline;
     if ipc2=chr(13) then
     begin;
      flag:=true;
      cset:=ipc1;
      clearline;
      calsetfn:='calset ';
      calsetfn[7]:=cset;
      reset(datafile {,calsetfn} );
      read(datafile,m1.mag,m1.ph,w1.re,w1.im);
      read(datafile,m2.mag,m2.ph,w2.re,w2.im);
      read(datafile,m3.mag,m3.ph,w3.re,w3.im);
     end;
    end;
   end;

   'Q','q': flag:=true;
   end; {case}

 until flag=true;

{Work out k}
  diff(w2,w1,wa);
  diff(z2,z3,za);
  diff(w2,w3,wb);
  diff(z2,z1,zb);
  mul(wa,za,num);
  mul(wb,zb,den);
  dvd(num,den,k);

{work out au}
  mul(k,z1,p);
  diff(z3,p,au);

{Work out bu}
  mul(p,w3,bu);
  mul(w1,z3,p);
  diff(bu,p,bu);

{Work out cu}
  unity.re:=1;
  unity.im:=0;
  diff(unity,k,cu);

{Work out du}
  mul(k,w3,p);
  diff(p,w1,du);

{Normalise}
  dvd(au,du,a);
  dvd(bu,du,b);
  dvd(cu,du,c);


  standards(a,b,c,d1,d2,d3,m1,m2,m3,ckit,cset);

  repeat {B option loop}
   repeat
    inverse;
    write('M=calculate z from measured w, S=standards ');
    not_inverse;
   repeat
     getkey(ipc);
   until ipc in ['S','s','M','m'];
    clearline;
    if ipc in ['M','m'] then zfind(a,b,c);
  until ipc in ['S','s'];
    standards(a,b,c,d1,d2,d3,m1,m2,m3,ckit,cset);
   repeat
     inverse;
    write('K=new CalKit  S=new CalSet  B=back  1=save CalKit  2=save CalSet ');
     not_inverse;
    repeat
      getkey(ipc);
    until ipc in ['K','k','S','s','B','b','1','2'];
     clearline;

     if ipc in ['1','2'] then
     begin;
      if ipc='1' then nm:='Kit' else nm:='Set';
     repeat
       inverse;
       write('Save Cal',nm,' number? ');
       not_inverse;
      repeat
       getkey(ipc1);
      until ipc1 in ['1'..'9'];
       clearline;
       inverse;
       write('Save as Cal',nm,' ',ipc1,'  <Return> ');
       not_inverse;
       getkey(ipc2);
       clearline;
      until ipc2=chr(13);

       if ipc='1' then
       begin;
        sname:='Calkit ';
        ckit:=ipc1;
       end
       else
       begin;
        cset:=ipc1;
        sname:='Calset ';
       end;

       sname[7]:=ipc1;
       rewrite(datafile {,sname} );
       if ipc='1' then
       begin;
       write(datafile,d1.mag,d1.ph,z1.re,z1.im);
       write(datafile,d2.mag,d2.ph,z2.re,z2.im);
       write(datafile,d3.mag,d3.ph,z3.re,z3.im);
      end
      else
      begin;
       write(datafile,m1.mag,m1.ph,w1.re,w1.im);
       write(datafile,m2.mag,m2.ph,w2.re,w2.im);
       write(datafile,m3.mag,m3.ph,w3.re,w3.im);
      end;
     end;

   until not(ipc in ['1','2']);
  until not(ipc in ['B','b']);
 until ipc in ['K','k'];
until false;
end.
