/*
 * file		as.c
 * date		9/6/85
 * author	A. Brouwer
 *
 * assemble 65C02 source code
 */

/*									*/
/*	stdio.h - definitions typically needed by a C program		*/
/*									*/
/*		FLEX version.						*/
/*									*/
/*	Most files compiled should have the line			*/
/*		#include "stdio.h"					*/
/*	in them to define these often used symbol constants.		*/
/*	It seems much better to use ERROR than -1!			*/
/*									*/

/*	Operating system definition					*/

#define FLEX	1

/* MAXLINE is the maximum text line size used in the standard library	*/
/* for the line processing functions (fgets(), fprintf(), printf() etc.)*/
/* If MAXLINE as defined is too small for your processing, you can	*/
/* change it but you MUST recompile ALL the functions in the standard	*/
/* library that need it AND replace them in the "stdlib.lib" file.	*/

#define MAXLINE 256		/* maxumum line buffer size		*/

#define BITS	short		/* set of sixteen bits			*/
#define BOOL	char		/* boolean (TRUE/FALSE) 		*/
#define BYTES	unsigned int	/* for address pointers and indexing	*/
#define BYTMASK 0xff		/* mask for low byte			*/
#define COUNT	short		/* small counting variable		*/
#define EOF	-1		/* end of file marker			*/
#define EOLN	10		/* end of line				*/
#define ERROR	-1		/* used by many functions as error ind. */
#define FALSE	0		/* BOOL false value			*/
#define FAST	register
#define FOREVER for(;;)
#define GLOBAL	extern
#define IMPORT	extern
#define LOCAL	static
#define INTERN	static
#define NO	0
#define NULL	0
#define OK	1
#define TEXT	char
#define TINY	char
#define TRUE	1		/* BOOL true value			*/
#define VOID	int		/* for functions returning nothing	*/
#define YES	1

/*									*/
/*	mem - direct memory accessing macro				*/
/*	this macro allows anything as its argument, coerces the 	*/
/*	argument into a pointer to character, and uses the pointer to	*/
/*	access a memory address directly.  The main purpose of this	*/
/*	macro is to cause th compiler to generate extended addressing	*/
/*	instructions to load or store fixed variables and i/o ports.	*/
/*	for example to directly access a 6850 ACIA at address $E000:	*/
/*									*/
/*		#define ACIA	0xE000					*/
/*									*/
/*	and later on in the program:					*/
/*									*/
/*		while((*mem(ACIA) & 1) == 0);	/* wait for byte      */*/
/*		ch=*mem(ACIA+1);		/* read ACIA data byte*/*/
/*									*/
/*	This sequence will generate instructions to directly access the */
/*	ACIA port.  The macro can also be an lvalue in an expression,	*/
/*	that is, you can assign values to the addressed byte:		*/
/*									*/
/*		*mem(ACIA)=3;	/* reset ACIA			      */*/
/*									*/

#define mem(adr) ((char *)(adr))

/*									*/
/*	character testing and conversion				*/
/*	The following macros are commented out so that the library	*/
/*	routines will be used instead.	This will make smaller programs.*/
/*	These macros should be un-commented if: 			*/
/*		1. you want your program to run faster			*/
/*		   (the compiler will optimize these expressions	*/
/*		    especially if they occur in the decision portion of */
/*		    a decision statement)				*/
/*		2. you are using a future version of this compiler that */
/*		   supports longs, floats, and doubles and you want	*/
/*		   these "functions" to be good for all data types.	*/
/*									*/
/*									*/

/***	*** commented out

#define isalpha(ch) (isupper(ch) || islower(ch))
#define isupper(ch) ((ch) >= 'A' && (ch) <= 'Z')
#define islower(ch) ((ch) >= 'a' && (ch) <= 'z')
#define isdigit(ch) ((ch) >= '0' && (ch) <= '9')
#define isspace(ch) ((ch) == ' ' || (ch) == '\n' || (ch) == '\t')
#define toupper(ch) (islower(ch) ? (ch)-('a'-'A') : ch)
#define tolower(ch) (isupper(ch) ? (ch)+('a'-'A') : ch)

***/

/*	file i/o definitions						*/

typedef int *FILE;		/* simplified declaration for file ptr. */

#define STDIN	1
#define STDOUT	2
#define STDERR	3

#define stdin	STDIN
#define stdout	STDOUT
#define stderr	STDERR

#define errno	(*mem(0xcc20))	/* place to find FMS error codes	*/

/* jump definition for non-local goto (see setjmp and longjmp)		*/

typedef struct jmp_buf
       {
       char    *sp;
       char    *pc;
       } jmp_buf;

/*									*/
/*	definitions for standard library routines			*/
/*	(only those routines that return non-integer type values	*/
/*	need to be defined)						*/
/*									*/

long	atol(); 	/* convert string to long			*/
float	atof(); 	/* convert string to float			*/
long	ulmul();	/* unsigned long multiply			*/
long	uldiv();	/* unsigned long divide 			*/
long	ulmod();	/* unsigned long modulus			*/
long	lseek();	/* seek to a file position			*/

/* end file stdio.h */

/*
 * file		asdef.c		include file for as
 *
 * defines and extern variables
 */

#define IF 0
#define ELSEIF 1
#define ELSE 2
#define ENDIF 3
#define DEFMAC 4
#define ENDM 5
#define LIB 6
#define CONT 7
#define ORG 8
#define END 9
#define EQU 10
#define SET 11
#define RMB 12
#define FCB 14
#define FDB 15
#define FCC 16
#define OPT 17
#define PAG 18
#define TTL 19
#define ARGM 20
#define ARGL 21
#define IFB 22
#define EXITM 23

#define MACRO -2
#define NOOPCODE -3

#define LIST 1
#define WARNING 64
#define LSYM 128

#define LPAG 2
#define LGEN 4
#define LMAC 8
#define LEXP 16
#define LCON 32
#define LMNE 64
#define LERR 128

#define PHASEDETECT	1
#define MNEMONIC	2
#define OPRMISSING	3
#define ADRMODE		4
#define MISSINGRSH	5
#define BRTOOLONG	6
#define TRUNCATE	7
#define LIBRARY		8

#define MACMISSING	10
#define MACNOTDEF	11
#define MACMULTIPLY	12
#define MACREPLACE	13
#define NOTALLOWED	14

#define EXPRESSION	15
#define MISSINGRH	16
#define DIGRANGE	17
#define LABNOTDEF	18
#define LABNOTFOUND	19
#define LABMISSING	20
#define LABINVALID	21
#define LABMULTIPLY	22
#define CONDITIONAL	23

#define MAXLIB	3
#define MAXIF	10
#define MAXMACRO 20
#define MAXOP	200
#define MAXLABEL 10
#define MAXBYT	100
#define MAXLOCAL 200
#define MAXMACLBUF 400

struct globallb {		/* global structure */
    struct globallb *left;	/* lower link */
    struct globallb *right;	/* higher link */
    unsigned value;		/* value */
    char mode;			/* mode */
    char name[0];
    };

struct locallb {
    struct locallb *link;
    struct localentry
	{
	unsigned value;
	int	label;
	} e[MAXLOCAL];
    };

struct macrocb {
    struct macrocb *link;
    char mode;
    struct macroline *l;
    struct macroline *p;
    char name[0];
    };
struct macroline {
    struct macroline *link;
    char line[0];
    };

char	lchar, *lpoin, *lendp;
char	*lblp, *opcp, *oprp, *comp;

char	ifmask;
int	opcode, opmode;
unsigned operand;
char	oprflag;

FILE	*fp;
char	*fname;
char	pass;
char	endflag;

char	oprf[4];
unsigned opr[4];

int	liblevel;
int	liblinecnt[MAXLIB];
FILE	*libfp[MAXLIB];
char	libfname[MAXLIB][20];

int	iflevel;
int	ififlevel;
char	ifbuf[MAXIF];

char	macdef;
int	maclevel;
struct	macrocb *argmac;
char	*argline;
struct	macrocb *macroroot;
struct	macrocb *macropoin;
struct	macrocb *macro[MAXMACRO];
int	mexline[MAXMACRO];
char	*macline[MAXMACRO];
char	maclbuf[MAXMACLBUF];

int	localcount;
struct	locallb *localroot;
struct	globallb *globalroot;
struct	globallb *globalpoin;

char	repmode;
char	strflag;
char	*macrpoin, *argrpoin;
char	labelmode;
unsigned labelvalue;
char	localflag;
unsigned locallow, localhigh;

char	listof;
char	listf;
int	linecut;

char	insbuf[10];
char	labbuf[40];
char	oprbuf[80];
char	line[100];	/* input string buffer */

char	bytcnt;
char	bytbuf[MAXBYT];

unsigned pc, pco;
int	linecount;
int	errcnt;
int	warcnt;
int	phasedet;

FILE	*outfp;
int	pagcnt;
int	paglmax, pagline;
char	pagttl[60];
char	errlcnt;
int	outlcnt;
char	outline[MAXLINE];

char	*binfname;
FILE	*binfp;
unsigned objloc, binbeg;
char	bincnt;
char	binbuf[256];

/* end file asdef.c */

main(argc,argv)
char **argv;
{
    int i,j;
    char *s;
    char c;
    outfp = STDOUT;
    fname = binfname = 0;
    argline = "";
    listof = 0;
    linecut = MAXLINE-5;
    while (--argc)
    {
	if (*(s = *++argv) == '-')
	{
	    ++s; c = 1;
	    while (c && *s)
	    switch (tolower(*s++))
	    {
	    case 's': listof |= LSYM; break;
	    case 'l': listof |= LIST; break;
	    case 'w': listof |= WARNING; break;
	    case 'c': c = 0;
		if (!sscanf (s, "%d", &linecut)) linecut = 78; break;
	    case 'o': c = 0; binfname = s; break;
	    case 'a': c = 0; argline = s; break;
	    }
	}
	else fname = s;
    }
    if (! fname)
    {
	phelp (); exit (0);
    }

    if ((fp = fopen(fname, "r")) == ERROR)
	printfex ("as: can't open source file \"%s\"\n", fname);
    fclose (fp);

    binopen ();

    globalroot = localroot = macroroot = argmac = 0;
    pass = 1;
    fputs ("Pass 1...\n", STDERR);
    execpass ();
    pass = 2;
    fputs ("Pass 2...\n", STDERR);
    execpass ();

    binwrite ();	/* save object bytes */
    binclose ();

    if (!(listof&LIST)) oprintf ("Last assembled address: %04x\n", pc);
    else oputchar ('\n');
    oprintf ("%d error(s) detected", errcnt-warcnt);
    oprintf (warcnt? ", %d warning(s) reported\n": "\n", warcnt);
    nextpage ();

    if (listof&LSYM)
    {
	if (!(listf&LPAG)) oputchar ('\n');
	symbolprint();
	nextpage ();
    }
}
/* file		aspass.c
 * date		9/5/85
 * author	A. Brouwer
 *
 *
 * perform one pass
 *
 * macros execution
 * conditional assembly
 * pseudo opcodes
 * label installation
 * split source line
 */

/*
 * execpass	get source line
 *
 * input	source file or macro line
 * output	source line to procline ()
 * exit		if eof during macro define
 *
 *		open source main file
 *		read source lines from (library)file or macro buffer
 */

execpass ()
{
    int i;
    struct macroline *ln;

    listf = LIST|LPAG|LGEN|LMAC|LMNE;
    pagcnt = 0; pagline = 0; paglmax = 58;
    *pagttl = '\0';
    macdef = 0; maclevel = -1;
    liblevel = liblinecnt[0] = linecount = 0;
    endflag = phasedet = errcnt = warcnt = pc = 0;
    strflag = repmode = 0;
    iflevel = -1; ififlevel = 0;

    if ((libfp[0] = fopen (fname, "r")) == ERROR) exit (-1);
    strcpy (libfname[0], fname);
    for (;;)
    {
	if (maclevel>=0)
	{
	    ln = macro[maclevel]->l;
		/* skip processed macro lines */
	    for (i = 0; i < mexline[maclevel] && ln; i++) ln = ln->link;
	    if (ln)	/* not end of macro */
	    {
		strcpy (line, ln->line);
		mexline[maclevel]++;
		procline ();
	    }
	    else maclevel--;
	}
	else if (fgets (line, 100, libfp[liblevel]))
	{
	    linecount++;
	    liblinecnt[liblevel]++;
	    procline ();
	}
	else
	    endflag = 1;
	if (endflag)
	{
	    if (macdef) errex ("end of file while defining macro");
	    endflag = 0;
	    fclose (libfp[liblevel--]);
	    if (liblevel<0) break;
	}
    }
}

/*
 * procline	proces source line
 *
 * input	source line
 * output	listing and/or binary object file
 * exit		macro or if level too deep or macro calling line full
 * error	many
 *
 */

procline ()	/* scan line */
{
    char c;
    int i;
    pco = pc;
    ifmask = iflevel < 0? 1: ifbuf[iflevel]&1;
    opcode = NOOPCODE;
    bytcnt = opmode = 0;
    lendp = lpoin = line;
    while (*lendp)
	if (*lendp == '\n') *lendp = '\0';
	else lendp++;
    lblp = opcp = oprp = comp = lendp;
    lchar = *lpoin;
    lineinit ();
    if (ltestsc())			/* label */
    {
	lblp = lpoin;
	ltskip();
    }
    else if (lchar != ';') ltskip ();
    if (ltestsc())			/* opcode */
    {
	opcp = lpoin;
	if (ifmask || (listf&LCON))
	{
	    macrpoin = 0;
	    macreplace ();
	}
	for (i=0; ltest(); lnext ())
		if (i<9) insbuf[i++] = tolower(lchar);
	insbuf[i] = '\0';
	ltskip();
	inssearch ();
	if (opcode == ERROR && (macropoin = macrosearch(opcp)))
	{
	    opcode = MACRO;
	    if (macropoin->p) opmode = -2;
	}
    }
    switch (opmode)		/* operand / comment */
    {
    case 200:				/* EQU */
    case 201:				/* IF */
    case 202: setopcom (0); break;	/* LIB */
    case 203: setopcom (2); break;	/* MACRO */
    case 204: setopcom (1); break;	/* FCC */
    case 205: setopcom (3); break;	/* END */
    case 206: setopcom (4); break;	/* TTL */
    default: if (opmode >= 1 || opcode == ERROR) setopcom(0); /* instr */
	else if (!opmode) setopcom(3);	/* MACRO / no operand */
	else setopcom (1);		/* MACRO */
    }

    if (opmode >= 200 && !macdef)
    {
	switch (opcode)
	{
	case IF: if (!ifmask)
		{
		    ififlevel++;
		    linecon ();
		}
		else
		{
		    if (iflevel >= MAXIF-1) errex ("if level too deep");
		    setchklabel ();
		    strflag = 1;
		    if (setopr())
			ifbuf[++iflevel] = operand?1:0;
		    else erriff ();
		    strflag = 0;
		    linemne ();
		}
		return;
	case ELSEIF: if (ififlevel) linecon ();
		else
		{
		    strflag = 1;
		    if (iflevel >= 0 && !(ifbuf[iflevel]&2) && setopr ())
			ifbuf[iflevel] = ifbuf[iflevel]? 4: operand?1:0;
		    else erriff ();
		    strflag = 0;
		    linemne ();
		}
		return;
	case ELSE: if (ififlevel) linecon ();
		else
		{
		    if (iflevel >= 0 && !(ifbuf[iflevel]&2))
			ifbuf[iflevel] = ifbuf[iflevel]? 2: 3;
		    else erriff ();
		    linemne ();
		}
		return;
	case ENDIF: if (ififlevel)
		{
		    ififlevel--;
		    linecon ();
		}
		else
		{
		    if (iflevel >= 0) iflevel--;
		    else erriff ();
		    linemne ();
		}
		return;
	}
    }
    if (!ifmask)
    {
	linecon (); return;
    }
    if (macdef)
    {
	if (opmode >= 200)
	switch (opcode)
	{
	case ENDM: linechk (); macdef = 0; return;
	case ORG:
	case END:
	case LIB:
	case CONT:
	case ARGM:
	case ARGL:
	case DEFMAC: erradd(NOTALLOWED);
			linechk (); return;
	}
	macroladd(); linechk (); return;
    }
    if (opmode >= 200)
    {
	switch (opcode) /* pseudo instructions */
	{
	case ARGL: if (!*argline && setoprf())
		{
		    argline = asalloc(strlen(lpoin)+1);
		    strcpy (argline, lpoin);
		}
		linechk (); break;
	case ARGM:
	case DEFMAC: i = opcode;	/* save */
		if (maclevel>=0) erradd (NOTALLOWED);
		else
		{
		  setlabel();
		  macdef = 1;
		  macro[0] = 0;
		  if (!*labbuf) erradd (MACMISSING);
		  else
		  {
		    for (c=0; c < 9 && lblp[c]; c++)
				insbuf[c] = tolower(lblp[c]);
		    insbuf[c] = '\0';
		    inssearch();
		    if (opcode != ERROR) erradd (MACMULTIPLY);
		    else
		    {
			macro[0] = macroinstal ();
			if (i == ARGM) argmac = macro[0];
		    }
		  }
		  if (i == ARGM) macdef = 0;
		}
		linechk (); return;
	case ENDM: erradd (NOTALLOWED);
		linechk (); break;
	case IFB: if (maclevel < 0) erradd (NOTALLOWED);
		else
		{
		    if (setmopr() != 2) erropm();
		    else if (!oprf[1] && opr[0])
		    {
			mexline[maclevel] += opr[1];
			if ((int)opr[1] < 0) mexline[maclevel]--;
		    }
		}
		linemne ();
		break;
	case EXITM: if (maclevel < 0) erradd (NOTALLOWED);
		    else maclevel--;
		linemne ();
		break;
	case CONT: fclose (libfp[liblevel--]);
	case LIB: setoprf();
	    if (liblevel >= MAXLIB-1)
		errex ("library nesting too deep");
	    if ((fp = fopen (lpoin, "r")) != ERROR)
	    {
		libfp[++liblevel] = fp;
		strcpy (libfname[liblevel], lpoin);
		liblinecnt[liblevel] = 0;
	    }
	    else if (liblevel < 0)
		errex ("can't continue with next file");
	    else erradd (LIBRARY);
	    linemne (); break;
	case EQU:
	case SET: setopr();
		if (setlabel () && !isdigit(*labbuf))
		{
		    if (pass == 1 && opcode == EQU)
		    {
			labelvalue = operand;
			globalinstal (labbuf);
		    }
		    if (pass == 2 || opcode == SET)
		    {
			if (looklabel(0)) globalpoin->value = operand;
		    }
		}
		else erradd (LABINVALID);
		lineequ (); break;
	case ORG: if (setopr()) pc = operand;
		else errex ("invalid origin specified");
		phasedet = 0;
		lineequ (); break;
	case END: if (!liblevel)
		{
		    if (*oprp)
		    {
			if (setopr () && pass == 2) startwrite (operand);
			lineequ ();
		    }
		    else linemne ();
		}
		else linemne ();
		endflag = 1; break;
	case RMB: setchklabel ();
		if (!(i = setmopr())) erropm();
		else if (!oprf[0])
		{
		    if (i == 1) pc += opr[0];
		    else while ((int)opr[0]--) outpass (opr[1]);
		}
		linefcb (); break;
	case FCB:
	case FDB:
	case FCC: setchklabel ();
		setfcb ();
		linefcb (); break;
	case OPT: setopt ();
		linemne (); break;
	case PAG: if (*oprp)
		{
		    setopr();
		    if ((int)operand < 0)
		    {
			if (pagline-(int)operand > paglmax) nextpage();
		    }
		    else
		    {
			if (operand >= 10) paglmax = operand;
			linemne ();
		    }
		}
		else nextpage();
		break;
	case TTL: setoprf (); if (strlen(lpoin) >= 59) lpoin[59] = '\0';
		strcpy (pagttl, lpoin);
		linemne (); break;
	}
	return;
    }
    setchklabel();		/* 6502 instructions */
    if (opcode == ERROR)
    {
	erradd (MNEMONIC);
	outpass (0xea);
	outpass (0xea);
	outpass (0xea);
    }
    else if (opcode == MACRO)
    {
	if (maclevel >= MAXMACRO-1)
		errex ("macro nesting too deep");
	if (opmode) setmac();
	macline[maclevel+1] = maclevel>=0?
	     macline[maclevel]+strlen(macline[maclevel])+1:maclbuf;
	if (macline[maclevel+1]+strlen(oprp)+1-maclbuf >= MAXMACLBUF)
		errex ("macro line buffer overflow");
	strcpy (macline[maclevel+1], oprp);

	if (errlcnt || (listf&LMAC)) lineins();
	else if (*lblp && maclevel<0)
	{
		opcp = oprp = lendp; lineins();
	}
	macro[++maclevel] = macropoin;
	mexline[maclevel] = 0;
	return;
    }
    else if (opcode == NOOPCODE)
    {
	if (!*lblp) return linechk ();
    }
    else
    {
	switch (opranal ())
	{
	case 1: outpass (opcode); break;
	case 2: outpass (opcode); outpass (operand); break;
	case 3: outpass (opcode); outpass (operand&255);
		outpass (operand/256);
		break;
	}
    }
    lineins();
}

/*
 * setchklabel	instal or check local or global label if present
 *		in label field
 *
 * input	source, pc
 */

setchklabel ()
{
    if (setlabel())
    {
	if (pass == 1)
	{
	    labelvalue = pc;
	    if (isdigit (*labbuf)) localinstal (labbuf);
	    else globalinstal (labbuf);
	}
	else if (!isdigit(*labbuf)) looklabel (1);
    }
}

/*
 * setlabel	get, replace and instal global label
 *
 * input	label pointer, label value
 * error	invalid label
 *
 *		before scanning it tries to replace macro argument
 */

setlabel ()
{
    *labbuf = '\0';
    if (*lblp)
    {
	linit (lblp);
	macrpoin = 0;
	macreplace ();
	if (islabel (lchar))
	{
	    if (isdigit(lchar)) partob (labbuf);
	    else labtob (labbuf);
	    labelmode = 0;
	    while (lchar == ':' || lchar == '.')
	    {
		labelmode += 16; lnext ();
	    }
	}
	if (ltest ()) erradd (LABINVALID);
    }
    return *labbuf;
}

/*
 * looklabel	search for global label and test if phase error
 *
 * input	label buffer
 *		mode is nonzero for program labels (phase error)
 * return	label found
 * error	not defined
 *		multiple label
 *		phasing error
 */
 
looklabel (mode)
{
    if (globalpoin = globalsearch (labbuf))
    {
	if (labelmode&1) erradd (LABMULTIPLY);
	else if (mode && !phasedet && labelvalue != pc)
	{
	    erradd (PHASEDETECT);
	    phasedet = linecount;
	}
    }
    else erradd (LABNOTDEF);
    return globalpoin;
}

/*
 * setopcom	split rest of source line into operand and comment field
 *
 * entry	mode for splitting
 * input	source line
 * output	operand and comment pointer
 *
 *		mode = 0 operand (string allowed)
 *		       1 standard instruction operand
 *		       2 operand until whitespace
 *		       3 comment field
 *		       4 operand field
 */

setopcom (mode)
char mode;
{
    char c;
    if (ltestsc() && mode < 3)
    {
	oprp = lpoin;
	if (mode < 2)
	while ((c = lchar) > ' ')
	{
	    lnext ();
	    if (!mode)
	    {
		if (c == '\'') lbslash();
	    }
	    else
	    {
		switch (c)
		{
		case '\'':
		case '/':
		case '"':
		    while (lchar && lchar != c) lbslash();
		    lnext ();
		}
	    }
	}
	ltskip ();
    }
    if (lchar)
    {
	if (mode == 4 && ltestsc()) oprp = lpoin;
	else comp = lpoin;
    }
}
/*
 * file		asmacro.c
 * date		9/5/85
 * author	A. Brouwer
 *
 *
 * macro search, instal and argument replace
 */

/*
 * macrosearch	search opcode in macro buffer
 * 
 * input	macro name in opcp
 * return	pointer to macro cb found
 */

macrosearch ()
{
    struct macrocb *mac;
    mac = macroroot;
    while (mac)
	if (strcmp (opcp, mac->name)) mac = mac->link;
	else return mac;
    return NULL;
}

/*
 * macroinstal	instal new macro definition
 *
 * input	macro name in labbuf
 *		source line for macro arguments
 * return	macro cb pointer
 *
 *		set multiple define flag if macro name exists
 *		store macro arguments
 *		checks correct argument string
 *
 *		uses parsearch () to scan arguments
 *
 * bugs		also used for argm define storage
 */

macroinstal ()
{
    struct macrocb *mac, *macold;
    struct macroline *mline, *mlink;
    char i, j;
    mac = macroroot;
    while (mac)
    {
	macold = mac;
	if (!strcmp (mac->name, labbuf))
	{
	    if (pass == 1) mac->mode |= 1;
	    break;
	}
	mac = mac->link;
    }
    if (pass == 1 && !mac)
    {
	mac = asalloc (sizeof(*mac) + strlen(labbuf)+1);
	if (!macroroot) macroroot = mac;
	else macold->link = mac;
	mac->link = 0;
	mac->l = mac->p = 0;
	mac->mode = 0;
	strcpy (mac->name, labbuf);
    }
    if (!mac) erradd (MACNOTDEF);
    else if (mac->mode&1) erradd (MACMULTIPLY);
    if (oprp)
    {
	for (i=0; parsearch (oprp, i, oprbuf); i++)
	{
	    if (mac && pass == 1 && !(mac->mode&1))
	    {
		mline = asalloc (sizeof (*mline) + strlen (oprbuf) +1);
		if (i) mlink->link = mline;
		else mac->p = mline;
		strcpy (mline->line, oprbuf);
		mlink = mline;
	    }
	    if (*oprbuf) for (j=1; oprbuf[j]; j++)
		if (!islabel(oprbuf[j]))
		{
		    errexp ();
		    break;
		}
	}
    }
    return mac;
}

/*
 * macroladd	add line to macro during definition
 *
 * input	source line
 */

macroladd ()
{
    struct macroline *ln, *lold;
    int len;
    if (macro[0] && pass == 1 && !(macro[0]->mode&1))
    {
	ln = macro[0]->l;
	while (ln)
	{
	    lold = ln; ln = ln->link;
	}
	len = lconstr ();
	ln = asalloc (sizeof(*ln) + len+1);
	if (!macro[0]->l) macro[0]->l = ln;
	else lold->link = ln;
	ln->link = 0; strcpy (ln->line, &outline[24]);
    }
}
lconstr ()	/* construct contiguous source line to outline */
{
    char *s;
    s = &outline[24];
    strcpy (s, lblp); s += strlen(lblp);
    if (*opcp)
    {
	*s++ = ' ';
	strcpy (s, opcp); s += strlen(opcp);
    }
    if (*oprp)
    {
	*s++ = ' ';
	strcpy (s, oprp); s += strlen(oprp);
    }
    if (*comp)
    {
	*s++ = ' ';
	strcpy (s, comp);
    }
    return strlen (&outline[24]);
}

/*
 * macreplace	replace current argument with macro calling argument
 * 
 * input	pointer to current position source line
 * output	adjusted source line
 * return	length replaced string
 *
 *		a replace pointer is set to prevent multiple replace
 *		within a replaced argument
 */

macreplace ()			/* macro expand */
{
    char res;
    res = 0;
    if (maclevel>=0 && lpoin > macrpoin)
    {
	res = repsearch (macro[maclevel]->p, macline[maclevel]);
	if (res) macrpoin = &lpoin[res-1];
    }
    return res;
}

/*
 * argreplace	replace current argument with defined argm argument
 * 
 * i/o/r	same as macreplace
 */

argreplace ()			/* command line */
{
    char res;
    res = 0;
    if (argmac && lpoin > argrpoin)
    {
	res = repsearch (argmac->p, argline);
	if (res)
	{
	    argrpoin = &lpoin[res-1];
	    if (argrpoin > macrpoin) macrpoin = argrpoin;
	}
    }
    return res;
}

repsearch (lp, li)	/* search and replace macro arg with calling arg */
struct macroline *lp;
char *li;
{
    int i, len, len2;
    for (i=0; lp; lp = lp->link, i++)
    {
	if (lp->line[0] && (len = repcmp (lp->line, lpoin)))
	{
	    if (len2 = parsearch (li, i, oprbuf))
	    {
		repcopy (len, len2);
		return len2;
	    }
	    else erradd (MACREPLACE);
	    return 0;
	}
    }
    return 0;
}
repcmp (s, t)	/* compare replace strings */
char *s, *t;	/* returns length string compared if ok */
{
    char i;
    for (i=1; *s == *t && *s; i++, s++, t++);
    if (!*s && !islabel(*t)) return i;
    else return 0;
}
repcopy (n1, n2)	/* replace in line macro arg with calling arg */
{			/* uses outline for temp string */
    char *s, *t;
    strcpy (&outline[24], oprbuf);
    s = &outline[24+n2-1]; t = &lpoin[n1-1];
    while (t <= lendp) *s++ = *t++;
    t = lendp += n2-n1;
    while (t >= lpoin) *t-- = *--s;
    if (lpoin < comp) comp += n2-n1;
    if (lpoin < oprp) oprp += n2-n1;
    if (lpoin < opcp) opcp += n2-n1;
    if (lpoin < lblp) lblp += n2-n1;
    lchar = *lpoin;
}

/*
 * parsearch	search parameter n from s and copy to t
 *		keep track of string deliminators
 *
 * entry	string pointers and parameter
 * return	length output string
 *
 * bugs		assumes output string to be 80 char long
 */

parsearch (t, n, s)		/* search parameter n */
char *s, *t;
{
    char i;
    char c, term;
    i = 0;
    if (*t > ' ')
    for (;;)
    {
	if (*t > ' ')
	switch (term = *t++)
	{
	case ',': break;
	case '\'':
	case '/':
	case '"': if (!n && i < 78) s[i++] = term;
		while ((c = *t) && *t++ != term)
		{
		    if (!n && i < 78) s[i++] = c;
		    if (c == '\\' && *t)
		    {
			c = *t++;
			if (!n) s[i++] = c;
		    }
		}
		if (c == term && !n && i < 79) s[i++] = c;
		continue;
	case '\\': if (*t)
		{
		    if (!n && repmode == 2 && i < 79) s[i++] = term;
		    term = *t++;
		}
	default: if (!n && i < 79) s[i++] = term;
		continue;
	}
	if (!n-- || !*t) break;
    }
    s[i++] = '\0';
    return n<0? i: 0;
}
/*
 * file		aslabel.c
 * date		9/5/85
 * author	A. Brouwer
 *
 *
 * local / global label storage
 * search labels
 * print global symbol table
 */

/*
 * localsearch	look for the closest local label
 *
 * entry	local label string
 * input	pc (program counter)
 * output	locallow, localhigh and localflag
 * return	localflag (non-zero if found)
 *
 *		localflag bit 0 equal to pc
 *			      1 lower found
 *			      2 higher found
 *		locallow, localhigh are the closest values found
 *
 *		uses parcon () to convert string to integer
 */

localsearch (labname)
char *labname;
{
    struct locallb *loc;
    unsigned value;
    int i, dig;
    dig = parcon (labname, 10);
    localflag = 0;
    loc = localroot;
    while (loc)
    {
	i = loc->link? MAXLOCAL: localcount;
	while (i--)
	{
	    if (loc->e[i].label == dig)
	    {
		value = loc->e[i].value;
		if (pc == value) localflag |= 1;
		else if (pc > value &&
			(!(localflag & 2) || value > locallow))
		{
			locallow = value; localflag |= 2;
		}
		else if (pc < value &&
			(!(localflag & 4) || value < localhigh))
		{
			localhigh = value; localflag |= 4;
		}
	    }
	}
	loc = loc->link;
    }
    return localflag;
}

/*
 * localinstal	instal local label
 *
 * entry	label string
 * input	pc (program counter)
 * output	new entry to local label buffer
 *
 *		uses parcon () to make label an integer
 */

localinstal (labname)
char *labname;
{
    struct locallb *locnext, *loc;
    int dig;
    dig = parcon (labname, 10);
    if (loc = localroot) while (loc->link) loc = loc->link;
    if (!localroot || localcount >= MAXLOCAL)
    {
	locnext = asalloc (sizeof (*locnext));
	if (!localroot) localroot = locnext;
	else loc->link = locnext;
	locnext->link = 0;
	loc = locnext;
	localcount = 0;
    }
    loc->e[localcount].value = pc;
    loc->e[localcount++].label = dig;
}

/*
 * globalsearch search global label
 *
 * entry	label name
 * output	labelvalue, labelmode
 * return	label found
 */

globalsearch (labname)
char *labname;
{
    struct globallb *globalcur;
    int res;
    globalcur = globalroot;
    while (globalcur)
    {
	if ((res = strcmp (labname, globalcur->name)) > 0)
		globalcur = globalcur->right;
	else if (res < 0) globalcur = globalcur->left;
	else
	{
	    labelvalue = globalcur->value;
	    labelmode = globalcur->mode;
	    return globalcur;
	}
    }
    return NULL;
}

/*
 * globalinstal instal new global label
 *
 * entry	label name
 * input	labelvalue, labelmode
 * output	new entry
 * return	new entry, else NULL
 *
 *		mark label if one tries to redefine the label
 */

globalinstal (labname)
char *labname;
{
    int length;
    int res;
    struct globallb *globalcur, *globallast;
    length = sizeof (*globalcur) + strlen(labname)+1;
    globalcur = globalroot;
    while (globalcur)
    {
	globallast = globalcur;
	if ((res = strcmp (labname, globalcur->name)) > 0)
		globalcur = globalcur->right;
	else if (res < 0) globalcur = globalcur->left;
	else
	{
		globalcur->mode |= 1;
		return NULL;
	}
    }
    globalcur = asalloc (length);
    if (! globalroot) globalroot = globalcur;
    else if (res < 0) globallast->left = globalcur;
    else globallast->right = globalcur;
    globalcur->value = labelvalue;
    globalcur->mode = labelmode;
    globalcur->left = globalcur->right = 0;
    strcpy (globalcur->name, labname);
    return globalcur;
}

/*
 * symbolprint	print global labels in alphabetically order
 *
 *		uses labbuf for temporary space
 */

symbolprint ()
{
    int cnt, res;
    char *s;
    struct globallb *cur, *last;
    s = labbuf;
    oputs ("\t\t\t\tSymbol table\n\n");
    if (!(cur = globalroot)) return;
    while (cur)
    {
	last = cur; cur = cur->left;
    }
    cnt = 0;
    while (last)
    {
	oprintf ("%-10.9s%04x%c",
			last->name,last->value, cnt == 4? '\n':'\t');
	cnt = (cnt+1)%5;
	strcpy (s, last->name);
	strcpy (&s[strlen(s)], ".");
	cur = globalroot; last = 0;
	while (cur)
	{
	    if ((res = strcmp (s, cur->name)) <= 0) last = cur;
	    if (res < 0) cur = cur->left;
	    else if (res > 0) cur = cur->right;
	    else cur = 0;
	}
	if (cnt && !last) oputchar ('\n');
    }
}
/*
 * file		asopc.c
 * date		 9/4/85
 * author	A. Brouwer
 *
 *
 * 65C02 opcodes and search
 */

#define OPCCOUNT 91	/* adjust if instruction count changes */

struct opce {		/* instruction, opcode, operand mode */
    char *s;
    char o, m;
    } opcba [] = {
	"adc", 0x60, 4,
	"and", 0x20, 4,
	"argl", 0x15, 204,
	"argm", 0x14, 203,
	"asl", 0x06, 13,
	"bbr", 0x0f, 21,
	"bbs", 0x8f, 21,
	"bcc", 0x90, 1,
	"bcs", 0xb0, 1,
	"beq", 0xf0, 1,
	"bit", 0x24, 14,
	"bmi", 0x30, 1,
	"bne", 0xd0, 1,
	"bpl", 0x10, 1,
	"bra", 0x80, 1,
	"brk", 0x00, 0,
	"bvc", 0x50, 1,
	"bvs", 0x70, 1,
	"clc", 0x18, 0,
	"cld", 0xd8, 0,
	"cli", 0x58, 0,
	"clv", 0xb8, 0,
	"cmp", 0xc0, 4,
	"cont", 0x07, 202,
	"cpx", 0xe0, 10,
	"cpy", 0xc0, 11,
	"dec", 0xc6, 12,
	"dex", 0xca, 0,
	"dey", 0x88, 0,
	"else",0x02, 205,
	"elseif", 0x01, 204,
	"end", 0x09, 205,
	"endif", 0x03, 205,
	"endm", 0x05, 205,
	"eor", 0x40, 4,
	"equ", 0x0a, 200,
	"exitm", 0x17, 205,
	"fcb", 0x0e, 200,
	"fcc", 0x10, 204,
	"fdb", 0x0f, 200,
	"if", 0x00, 204,
	"ifb", 0x16, 201,
	"inc", 0xe6, 12,
	"inx", 0xe8, 0,
	"iny", 0xc8, 0,
	"jmp", 0x4c, 3,
	"jsr", 0x20, 2,
	"lda", 0xa0, 4,
	"ldx", 0xa2, 6,
	"ldy", 0xa0, 7,
	"lib", 0x06, 202,
	"lsr", 0x46, 13,
	"macro", 0x04, 203,
	"nop", 0xea, 0,
	"opt", 0x11, 204,
	"ora", 0x00, 4,
	"org", 0x08, 200,
	"pag", 0x12, 200,
	"pha", 0x48, 0,
	"php", 0x08, 0,
	"phx", 0xda, 0,
	"phy", 0x5a, 0,
	"pla", 0x68, 0,
	"plp", 0x28, 0,
	"plx", 0xfa, 0,
	"ply", 0x7a, 0,
	"res", 0x0c, 200,
	"rmb", 0x07, 20,
	"rol", 0x26, 13,
	"ror", 0x66, 13,
	"rti", 0x40, 0,
	"rts", 0x60, 0,
	"sbc", 0xe0, 4,
	"sec", 0x38, 0,
	"sed", 0xf8, 0,
	"sei", 0x78, 0,
	"set", 0x0b, 200,
	"smb", 0x87, 20,
	"sta", 0x80, 5,
	"stx", 0x82, 8,
	"sty", 0x80, 9,
	"stz", 0x64, 16,
	"tax", 0xaa, 0,
	"tay", 0xa8, 0,
	"trb", 0x14, 15,
	"tsb", 0x04, 15,
	"tsx", 0xba, 0,
	"ttl", 0x13, 206,
	"txa", 0x8a, 0,
	"txs", 0x9a, 0,
	"tya", 0x98, 0,
	};

/*
 * inssearch	search instruction
 *
 * input	insbuf instruction buffer
 * output	opcode (ERROR if not found) opmode
 * return	opcode
 */
 
inssearch ()
{
    int i,j,t;
    int res;
    i=0; j=OPCCOUNT-1;
    opcode = ERROR;
    while (i<=j)
	if ((res = strcmp (insbuf, opcba[t=(i+j)/2].s)) > 0) i = t+1;
	else if (res < 0) j = t-1;
	else
	{
	    opcode = opcba[t].o;
	    opmode = opcba[t].m;
	    break;
	}
    return opcode;
}
/*
 * file		asopr.c
 * date		9/5/85
 * author	A. Brouwer
 *
 *
 * operand scanning (instructions, field constants, macro call)
 * assembly option (re)set
 * frequently used error messages
 */

/*
 * opranal	instruction operand
 *
 * input	source line
 *		opcode, opmode
 * output	operand
 *		number of bytes
 * error	if incorrect addressing mode or
 *		incorrect closed operand
 * warning	truncated immediate or direct operand
 *
 * bugs		assumes 65C02 processor
 */

opranal ()		/* instruction operand */
{
    char imm, zadr, abs, ind, indx, indy, bitf;
    unsigned bitopr, memopr;

    imm = zadr = abs = ind = indx = indy = bitf = 0;
    if (!opmode) return 1;
    setoprf(); macreplace ();
					/* accu addressing */
    if ((opmode == 12 || opmode == 13) && tolower(lchar) == 'a' && !lpoin[1])
    {
	opcode ^= opmode == 12? 0xfc: 0x0c;		/* inc a asl a */
	return 1;
    }
    if (pass == 1)
    {
	switch (opmode)
	{
	case 1: return 2;
	case 2:
	case 3: return 3;
	}
    }
    switch (lchar)
    {
    case '#': imm = 1; lnext (); break;
    case '<': zadr = 1; lnext (); break;
    case '>': abs = 1; lnext (); break;
    case '[': ind = 1; lnext (); break;
    }
    scanopr ();
    if (lchar == ',')
    {
	if (opmode == 20 || opmode == 21)
	{
	    lnext (); bitopr = operand; bitf = 1;
	    scanopr ();
	    if (opmode == 21 && lchar == ',')
	    {
		lnext (); memopr = operand; bitf = 2;
		scanopr ();
	    }
	}
	else
	{
	    if (tolower (lnext()) == 'x') indx = 1;
	    else if (tolower (lchar) == 'y' && ! ind) indy = 1;
	    else erradr ();
	    lnext();
	}
    }
    if (ind)
    {
	if (lchar == ']')
	{
	    if (lnext() == ',')
	    {
		if (tolower (lnext()) == 'y' && ! indx) indy = 1;
		else erradr ();
		lnext ();
	    }
	}
	else erradd (MISSINGRSH);
    }
    if (ltest())
    {
	oprflag = 1;
	errexp ();
    }
    switch (opmode)
    {
    case 1: if (imm|ind|indx|indy|zadr|abs) erradr ();	/* beq bpl ... */
	    if (!oprflag) operand -= pc+2;
	    if ((int)operand >= 128 || (int)operand < -128)
	    {
		operand = 0;
		erradd (BRTOOLONG);
	    }
	    return 2;
    case 2: if (imm|ind|indx|indy|zadr|abs) erradr();	/* jsr */
	    return 3;
    case 3: if ((imm|indy|zadr|abs) || !ind && indx) erradr (); /* jmp */
	    if (ind) opcode |= 0x20;
	    if (indx&ind) opcode |= 0x10;
	    return 3;
    case 4: if (imm)					/* lda ora ... */
	    {
		opcode |= 0x09; valchk ();
		return 2;
	    }
    case 5: if (imm) erradr ();				/* sta */
	    if (ind)
	    {
		if (indy) opcode |= 0x11;
		else if (indx) opcode |= 0x01;
		else opcode |= 0x12;
		valchk (); return 2;
	    }
	    if (indy)
	    {
		if (zadr) erradr ();
		opcode |= 0x19;
		return 3;
	    }
	    if (indx) opcode |= 0x10;
	    if (zadr || !oprflag && !abs && operand < 256)
	    {
		opcode |= 0x05; valchk (); return 2;
	    }
	    opcode |= 0x0d;
	    return 3;
    case 6:						/* ldx */
    case 7:						/* ldy */
    case 10:						/* cpx */
    case 11: if (imm)					/* cpy */
	    {
		valchk (); return 2;
	    }
    case 8:						/* stx */
    case 9: if (imm|ind) erradr ();			/* sty */
	    if (indy || indx)
	    {
		if (opmode >= 10 || (indy & opmode) ||	/* cpx,y ldy,y */
		   opmode >= 8 && !zadr &&
			(abs || oprflag || operand > 256))/* stx,y */
			erradr ();
		else
		    opcode |= 0x10;
	    }	    
	    if (zadr || !oprflag && !abs && operand < 256)
	    {
		opcode |= 0x04; valchk (); return 2;
	    }
	    opcode |= 0x0c;
	    return 3;

    case 14: if (imm)					/* bit */
	    {
		opcode = 0x89; valchk (); return 2;
	    }
    case 12:						/* inc */
    case 13:						/* asl */
    case 15:						/* trb */
    case 16:						/* stz */
	    if (imm|ind|indy) erradr ();
	    if (indx)
	    {
		if (opmode != 15) opcode |= 0x10;
		else erradr ();
	    }
	    if (zadr || !oprflag && !abs && operand < 256)
	    {
		valchk (); return 2;
	    }
	    if (opcode != 16) opcode |= 0x08;
	    else opcode = indx? 0x9e: 0x9c;
	    return 3;
    case 20: if (!imm || ind || bitf != 1) erropm ();	/* rmb */
	    if (bitopr > 7) erradr ();
	    opcode |= (bitopr&7)*16;
	    valchk (); return 2;
    case 21: if (!imm || ind || bitf != 2) erropm ();	/* bbr */
	    if (bitopr > 7) erradr ();
	    opcode |= (bitopr&7)*16;
	    if (!oprflag) operand -= pc+3;
	    if ((int)operand >= 128 || (int)operand < -128)
	    {
		operand = 0;
		erradd (BRTOOLONG);
	    }
	    bitopr = operand; operand = memopr;
	    valchk ();
	    operand += 256 * bitopr;
	    return 3;
    }
}

/*
 * setfcb	byte, word and ascii field constants
 *
 * input	source line
 * output	operand, comment pointer
 *		object
 * error	if the operand is not correctly closed
 *
 *		multiple operands are allowed, each separated with ','
 *
 * bugs		the program counter is updated after each operand
 */

setfcb ()		/* byte word ascii field constants */
{
    char i;
    if (setoprf())
    {
	while (lchar)
	{
	    if (opcode == FCC) strflag = 1;
	    scanopr ();
	    if (opcode == FCC && strflag > 1)
		for (i=2; i<strflag; i++) outpass(oprbuf[i-2]);
	    else if (opcode == FDB)
	    {
		outpass (operand&255); outpass (operand/256);
	    }
	    else
	    {
		outpass (operand); valchk ();
	    }
	    strflag = 0;
	    if (lttest ()) break;
	}
    }
    else erropm ();
}

/*
 * setmac	define macro calling argument operand field
 *
 * input	source line
 * output	operand and comment pointer
 *
 *		arguments must be separated with ','
 *		uses scanopr () to replace if nessesary nested macro
 *		arguments
 */

setmac ()		/* set macro call operand field */
{
    if (maclevel >= 0)
    {
	setoprf();
	while (lchar)
	{
	    repmode = strflag = 1;
	    scanopr();
	    lbslash (); /* skip next char */
	    repmode = strflag = 0;
	}
    }
}

valchk ()		/* check if operand truncated */
{
    if (operand > 256) waradd (TRUNCATE);
}

outpass (c)		/* output byte to bytbuf and binfile */
char c;
{
    if (pass == 2)
    {
	if (bytcnt < MAXBYT) bytbuf[bytcnt++] = c;
	binbyt (c);
    }
    pc++;
}    

/*
 * setopt	(re)set assembly options
 *
 * input	source line
 * output	listf
 * error	not recognized option
 */

setopt ()
{
    int i, j;
    if (setoprf())
    {
	while (lchar)
	{
	    if (i = parsearch (lpoin, 0, oprbuf))
	    {
		linit (&lpoin[i-1]);
		if ((j = optcmp ()) == ERROR) errexp ();
		else listf = j&1? listf & ~(1 <<(j/2)):
					listf | 1 <<(j/2);
	    }
	    if (lttest ()) break;
	}
    }
    else erropm ();
}
optcmp()
{
    int i;
    for (i = 0; islabel(oprbuf[i]); i++)
		oprbuf[i] = tolower(oprbuf[i]);
    for (i = 0; parsearch 
("lis,nolis,pag,nopag,gen,nogen,mac,nomac,\
exp,noexp,con,nocon,mne,nomne,err,noerr",
	i, labbuf); i++)
    {
	if (repcmp(labbuf,oprbuf)) return i;
    }
    return ERROR;
}

lttest ()		/* operand termination ',' or \0 */
{			/* return zero if ok */
    if (lchar == ',') lnext();
    else if (lchar)
    {
	errexp ();
	return 1;
    }
    return 0;
}

setoprf ()	/* initialize souce line pointer to operand pointer */
{
    macrpoin = argrpoin = 0;
    return linit (oprp);
}

setopr ()		/* single operand */
{			/* return nonzero if ok */
    setoprf(); scanopr ();
    if (ltest())
    {
	oprflag = 1;
	errexp ();
    }
    return oprflag? 0 : 1;
}

setmopr ()		/* max 4 operands */
{			/* return number of operands */
    char i;
    for (i=0; i<4; i++) oprf[i] = (opr[i] = 0) + 1;
    i = 0;
    if (setoprf())
    {
	while (lchar && i<4)
	{
	    scanopr ();
	    opr[i] = operand;
	    oprf[i++] = oprflag;
	    if (lttest ()) break;
	}
    }
    return i;
}

/*
 * error messages
 */

erropm ()
{
    erradd (OPRMISSING);
}
errexp ()
{
    erradd (EXPRESSION);
}
erradr ()
{
    erradd (ADRMODE);
}
erriff ()
{
    erradd (CONDITIONAL);
}
/*
 * file		asscan.c
 * date		9/5/85
 * author	A. Brouwer
 *
 *
 * operand scanning allowing arithmetic, logical and relational operations
 * move source line to buffer (parameter, label)
 * convert string to integer according base
 * test ishex, islabel
 */

/*
 * scanopr	calculate next operand
 *
 * input	source line
 *		strflag, repmode
 * output	oprflag, operand
 *		strflag, oprbuf (string)
 *
 * error	if no or incorrect format encountered
 *		it does not check if the operand is correctly ended
 *
 *		At the highest level, it tries to replace the current
 *		argument with an argl argument or with a macro calling
 *		argument.
 *
 *		If strflag is nonzero, string operations are allowed.
 *		On return strflag greater than one (1) indicates the
 *		size+1 of the string positioned to oprbuf.
 *		For string operations only "!=" and "==" are correctly
 *		handled. The other operations ignore the string.
 *		If using the "!=" or "==" string operation, the replace
 *		macro argument must be the first argument. Otherwise
 *		results are unpredictable (The macro replace also uses
 *		oprbuf for temporary space).
 *
 *		If repmode is nonzero, argument scanning is performed to
 *		replace nested macro arguments. In this case no error
 *		messages are printed and no argm argument replace is done.
 */

scanopr ()		/* scan operand value */
{
    unsigned data;
    oprflag = data = 0;
    if (ltest())
	scanqassign (&data);
    else if (!repmode)
    {
	oprflag = 1;
	erropm ();
    }
    operand = data;
}
scanqassign (val)			/* ? : */
unsigned *val;
{
    scancor (val);
    if (lchar == '?')
    {
	lnext(); scanassign (val);
    }
}
scanassign (val)			/* ? : */
unsigned *val;
{
    unsigned data;
    unsigned u;
    scancor (&data);
    while (lchar == '?')
    {
	lnext(); scanassign(&data);
    }
    if (lchar == ':')
    {
	if (*val) u = data;
	lnext(); scancor(&data);
	while (lchar == '?')
	{
	    lnext(); scanassign(&data);
	}
	if (!*val) u = data;
    }
    else if (!repmode)
    {
	errexp ();
	oprflag |= 2;
    }
    *val = u;
}
scancor (val)		/* || */
unsigned *val;
{
    unsigned data;
    scancand (val);
    while (lchar == '|' && lpoin[1] == '|')
    {
	lnext(); lnext();
	scancand (&data);
	*val = *val || data;
    }
}
scancand (val)		/* && */
unsigned *val;
{
    unsigned data;
    scanor (val);
    while (lchar == '&' && lpoin[1] == '&')
    {
	lnext(); lnext();
	scanor (&data);
	*val = *val && data;
    }
}
scanor (val)		/* | */
unsigned *val;
{
    unsigned data;
    scaneor (val);
    while (lchar == '|')
    {
	lnext ();
	scaneor (&data);
	*val |= data;
    }
}
scaneor (val)		/* ^ */
unsigned *val;
{
    unsigned data;
    scanand (val);
    while (lchar == '^')
    {
	lnext ();
	scanand (&data);
	*val ^= data;
    }
}
scanand (val)		/* & */
unsigned *val;
{
    unsigned data;
    scanequal (val);
    while (lchar == '&')
    {
	lnext ();
	scanequal (&data);
	*val &= data;
    }
}
scanequal (val)		/* == != */
unsigned *val;
{
    unsigned data;
    char datoper, datop2;
    scanlsgt (val);
    while ((lchar == '!' || lchar == '=') && lpoin[1] == '=')
    {
	datoper = lchar;
	datop2 = strflag;
	lnext (); lnext();
	scanlsgt (&data);
	*val = ((datop2 >= 2 && strflag > datop2) ?
		strcmp (oprbuf, &oprbuf[datop2-1]) == 0:
		*val == data) ^ datoper == '!';
    }
}
scanlsgt (val)		/* < <= > >= */
unsigned *val;
{
    unsigned data;
    char datoper, datop2;
    scanshift (val);
    while (lchar == '>' || lchar == '<')
    {
	datoper = lchar;
	if ((datop2 = lnext ()) == '=') lnext();
	scanshift (&data);
	*val = (datop2 == '=' && *val == data) ||
		(datoper == '<' && *val < data) ||
		(datoper == '>' && *val > data);
    }
}
scanshift (val)		/* << >> */
unsigned *val;
{
    unsigned data;
    char datoper;
    scanadsub (val);
    while (lchar == '>' && (datoper = lpoin[1]) == '>' ||
		lchar == '<' && datoper == '<')
    {
	lnext (); lnext ();
	scanadsub (&data);
	if (datoper == '<') *val <<= data;
	else *val >>= data;
    }
}
scanadsub (val)		/* + - */
unsigned *val;
{
    unsigned data;
    char datoper;
    scanmudimo (val);
    while (lchar == '+' || lchar == '-')
    {
	datoper = lchar;
	lnext ();
	scanmudimo (&data);
	if (datoper == '+') *val += data;
	else *val -= data;
    }
}
scanmudimo (val)	/* * / % */
unsigned *val;
{
    unsigned data;
    char datoper;
    scanpar (val);
    while (lchar == '*' || lchar == '/' || lchar == '%')
    {
	datoper = lchar;
	lnext ();
	scanpar (&data);
	switch (datoper)
	{
	case '*': *val *= data; break;
	case '/': *val /= data; break;
	case '%': *val %= data; break;
	}
    }
}
scanpar (val)		/* scan parameter */
unsigned *val;
{
    char base;
    *val = 0;
    if ((base = macreplace ()) && repmode)
    {
	linit (&lpoin[base-1]);
	return;
    }
    if (!repmode) argreplace ();
    if (strflag >= 1)
    {
	switch (lchar)
	{
	case '/':
	case '\'':
	case '"': base = lchar; lnext();
		while (lchar && lchar != base && strflag < 80)
			oprbuf[-1+strflag++] = lbslash();
		oprbuf[-1+strflag++] = '\0';
		if (lchar == base) lnext();
		else if (!repmode)
		{
		    errexp ();
		    oprflag |= 1;
		}
		return;
	}
    }
    base = 0;
    switch (lchar)
    {
    case '(':	lnext ();
		scanqassign (val);
		if (lchar != ')') erradd (MISSINGRH);
		else lnext ();
		return;
    case '!':	lnext ();
		scanpar (val);
		*val = *val? 0: 1;
		return;
    case '-':	lnext ();
		scanpar (val);
		*val = -*val;
		return;
    case '~':	lnext ();
		scanpar (val);
		*val = ~*val;
		return;
    case '\'':	if (!lnext()) oprflag |= 1;
		*val = lbslash();
		if (lchar == '\'') lnext();
		return;
    case '*':	lnext ();	/* program counter */
		*val = pc;
		return;
    case '$':	lnext ();
		if (ishex (lchar))
		{
		    base = 16; break;
		}
		*val = pc; /* program counter */
		return;
    case '%':	lnext (); base = 2; break;
    case '@':	lnext (); base = 8; break;
    }
    if (base) partob (oprbuf);
    if (! base && isdigit (lchar))
    {
	partob (oprbuf);
	switch (tolower (lchar))
	{
	case '.': if (!repmode) localsearch (oprbuf);
		switch (tolower (lnext ()))
		{
		case 'f':if (localflag&4)
			    *val = localhigh;
			else base = 1;
			lnext (); break;
		case 'b':if (localflag&2)
			    *val = locallow;
			else base = 1;
			lnext (); break;
		default: switch (localflag)
			{
			case 1: *val = pc; break;
			case 2:
			case 3: *val = locallow; break;
			case 4:
			case 5: *val = localhigh; break;
			case 6:
			case 7: *val = 
				pc - locallow < localhigh - pc?
					locallow: localhigh;
				break;
			default: base = 1;
			}
		}
		if (base && !repmode)
		{
		    oprflag |= 2; erradd (LABNOTFOUND);
		}
		return;
	case 'h': lnext (); base = 16; break;
	case 'o': lnext