6502 PC/AT keyboard interface V0.1                1-Jan-87  0:00             page 0


		;
		;files		pckey.MAC
		;
		;last modified	11th May 2001
		;
		;program	PCKEY
		;
		;function	Routines for reading from
		;               and sending commands to PC/AT keyboard
		;               interface thru' spare 6522 VIA at $E160
		;               Data input connected to PB0
		;               Keyboard clock connected to PB1
		;
		;usage		PCKEY
		;
		;by		Andrew Gregory
		;		22, St. Mary's Avenue
		;		Teddington
		;		Middx.
		;		TW11 0HZ
		;		England
		;
		;zero page workspace needed in pckeyhd.mac
		;
	  0080		org	$80
0080		LEDBits	res	1
0081		KeyBits	res	1
0082		RxChar	res	1
0083		TxChar	res	1
0084		RxParity res	1
0085		TxParity res	1
		;
	  A000		org	$A000
A000 4C 2AA1		jmp	begin

		;
		; Libraries
		; IBM AT keyboard interface - low level hardware routines for
		; sending and receiving bytes.
		;
		; File PCKEYHD.MAC
		;
		; Created: 10th May 2001
		;
		; From program by L Davison on web.
		;
		; Extracted rouyines adapted to DOS65, A.P.Gregory, May 2001
		;
		; Uses port B of 6522 VIA, bit 0 is data, bit 1 is clock.
		;
		; GetAT		scans the keyboard and waits for a key
		; ScanAT		scans the keyboard, returns after timeout if no key
		; SendAT		sends a byte to the keyboard
		;
		; The GetAT and ScanAT routines automatically instruct the keyboard to
		; re-send if parity error (by transmitting $FE) then re-read the character.
		;
		; The SendAT reads the keyboard acknowledgement character, and return this
		; in register A. If the acknowledgement character is $FE this indicates a
		; parity (or other) error, and so SendAT re-sends the character.
		;

		; addresses used ...
6502 PC/AT keyboard interface V0.1                1-Jan-87  0:00             page 1

	  E160	VIA	equ	$E160		; type 6522 VIA
	  E160	ORB	equ	VIA		; input/output, bit 0 is data
		; bit 1 is clock, bits 2 to 7 are unused
	  E162	DDRB	equ	VIA+2		; Data direction register
		; Logic 1 indicates output line.

		; Note: Pull-up resisters (4k7) are needed on both lines to  enable reading
		; of keyboard.
		;
		; scan the keyboard to see if a key is waiting. returns scancode in A
		; and RxChar, or $00 if no key waiting. this is the main entry point.
		; On exit, the keyboard is left in standby.
		;
		; if the keyboard response seems flakey when holding a key for auto
		; repeat then increase the timeout, it helps. This was the shortest
		; I could reliably get away with with all the test keyboards.
		; possible bytes (apart from scancodes) are ...

		; $AA		Power On Self Test Passed (BAT Completed)
		; $EE		See Echo Command (Host Commands)
		; $FA		Acknowledge
		; $FE		Resend
		; $FF		Error or Buffer Overflow

A003		ScanAT
		;	LDX	#$0D		; set timeout count (150uS, 1MHz 6502)
A003 A2 1A		LDX	#$1A		; set timeout count (150uS, 2MHz 6502)
A005 A9 00		LDA	#$00
A007 8D 62E1		STA	DDRB		; both lines now inputs
A00A A9 02	WaitW0	LDA	#$02		; set for clock line bit test
A00C 2D 60E1		AND	ORB		; test the clock line
A00F F0 1A		BEQ	WaitW1		; go do rest if the clock falls
A011 CA			DEX			; else decrement timeout value
A012 D0 F6		BNE	WaitW0		; go wait some more if time is not up yet
A014 86 82		STX	RxChar		; store zero in buffer

		; Put into standby, typed characters will be stored in buffer in keyboard
A016		StandBy
A016 A9 02		LDA	#2		; inhibit activity by  taking clock low
A018 8D 62E1		STA	DDRB		; clock line now output
A01B A9 00		LDA	#0
A01D 8E 60E1		STX	ORB		; take clock pin low, data floats high
A020 60			RTS			; return with A=0, Z=1


		; Rx error, try again
A021		RxWasNOK
A021 A9 FE		LDA	#$FE		; resend command
A023 20 6AA0		JSR	SendAT		; send A to keyboard

		; scan the keyboard waitfor a key. returns scancode in A and RxChar.
		; this is the secondary entry point.
		; CLI issued before return
		; SEI can be given before calling for max. reliability
		;
A026		GetAT
A026 A9 00		LDA	#0		; clock high, data high
A028 8D 62E1		STA	DDRB		; both lines now  inputs (floating high)
A02B		WaitW1
A02B 20 54A0		JSR	GetBit		; get bit from keyboard (start bit)
		; note that SEI set in GetBit
		; but only once activity detected

A02E A2 08		LDX	#8		; eight data bits to get
6502 PC/AT keyboard interface V0.1                1-Jan-87  0:00             page 2

A030 A0 01		LDY	#1		; 1's count to one (odd parity)
A032		RecByte
A032 20 54A0		JSR	GetBit		; get bit from keyboard (data bit)
A035 90 01		BCC	NoRx1		; brabch if bit was zero
A037 C8			INY			; else increment 1's count
A038		NoRx1
A038 66 82		ROR	RxChar		; bit into receive byte
A03A CA			DEX			; decrement bit count
A03B D0 F5		BNE	RecByte		; loop if more data
A03D 20 54A0		JSR	GetBit		; get bit from keyboard (parity bit)
A040 26 84		ROL	RxParity	; bit into parity byte b0
A042 20 54A0		JSR	GetBit		; get bit from keyboard (stop bit)
A045 58			CLI
A046 20 16A0		JSR	StandBy

A049 98			TYA			; get computed parity
A04A 45 84		EOR	RxParity	; compare with received parity bit
A04C 6A			RORA			; only interested in bit 0
A04D B0 D2		BCS	RxWasNOK	; go try again if not ok
		; (rx parity <> computed parity)

A04F A5 82		LDA	RxChar		; else copy scancode
A051 F0 15		BEQ	ClearOF		; branch if was error or buffer overflow

A053 60			RTS

		; gets a bit from the keyboard, returned in carry bit
A054		GetBit
A054 A9 02		LDA	#2
A056 2D 60E1		AND	ORB		; test the clock line
A059 D0 F9		BNE	GetBit		; wait for the clock to fall
A05B 78			SEI
A05C AD 60E1		LDA	ORB		; get data
A05F 4A			LSRA			; data bit to Cb
A060 A9 02	WaitR1	LDA	#2		; set for clock line bit test
A062 2D 60E1		AND	ORB		; test the clock line
A065 F0 F9		BEQ	WaitR1		; wait for the clock to rise

A067 60			RTS



A068 A9 F4	ClearOF	LDA	#$F4		; Clear if keybd buffer overflow


		; SendAT sends the special codes to the keyboard, codes are ..
		; $ED		set the LEDs according to the next byte I send
		;		bit 0 = scroll lock
		;		bit 1 = num lock
		;		bit 2 = caps lock
		;		bits 3-7 must be 0, 1 = LED on
		; $EE		echo, keyboard will respond with $EE
		; $F0		set scan code set, upon sending the keyboard will
		;		respond with ACK ($FA) and then wait for a second
		;		byte. sending $01 to $03 determines the code set
		;		used, sending $00 will return the code set currently
		;		in use.
		; $F3		set typematic repeat rate, upon sending the keyboard
		;		will respond with ACK ($FA) and then wait for a second
		;		byte. this byte sets the rate.
		; $F4		keyboard enable, clears the keyboard buffer and starts
		;		scanning.
		; $F5		keyboard disable, clears the keyboard buffer and stops
		;		scanning.
6502 PC/AT keyboard interface V0.1                1-Jan-87  0:00             page 3

		; $F6		restore default values upon sending the keyboard will
		;		respond with ACK ($FA)
		; $FE		retransmit the last character please upon sending the
		;		keyboard will respond by resending the last character
		; $FF 		reset keyboard
		;
A06A		SendAT				; entry point putting in Standby
A06A 48			PHA
A06B 20 16A0		JSR	StandBy
A06E 68			PLA
A06F A2 20		LDX	#$20		; wait at least 60us
A071 CA		L	DEX
A072 D0 FD		BNE	L

A074		SendAT2				; Entry point if in Standby
A074 85 83		STA	TxChar		; save A in transmit buffer

A076		SendAT3				; entry point for re-sending last char
A076 78			SEI			; disable interrupts
A077 A2 08		LDX	#$08		; eight bits to send
A079 A0 01		LDY	#1
A07B 8C 62E1		STY	DDRB		; data low, clock high
A07E 88			DEY
A07F 8C 60E1		STY	ORB		; Now clock pin input, data stays output
		; Y=1
		; clock pin input now floating high

A082		SendByte
A082 66 83		ROR	TxChar		; bit into carry
A084 90 01		BCC	NotOne		; skip increment if zero
A086 C8			INY			; else increment 1's count
A087		NotOne
A087 20 BDA0		JSR	SendBit		; send bit to keyboard
A08A CA			DEX			; decrement bit count
A08B D0 F5		BNE	SendByte	; loop if not all done

A08D 66 83		ROR	TxChar		; last bit back into TxChar
A08F 98			TYA			; copy parity count
A090 4A			LSRA			; parity bit into carry
A091 20 BDA0		JSR	SendBit		; send parity bit to keyboard

A094 A9 00		LDA	#0
A096 8D 62E1		STA	DDRB		; data now input (as well as clk

A099 A9 01		LDA	#1
A09B 2D 60E1	p3	AND	ORB		; keyboard acknowledgement if data low
A09E D0 FB		BNE	p3
A0A0 58			CLI			; allow interrupts

		; keyboard should stop generating clock soon, but wait
		; for a while after last cycle just in case
A0A1 A9 02	p1	LDA	#2
A0A3 A2 14		LDX	#20
A0A5 2D 60E1	p2	AND	ORB		; Clock low (i.e. being generated)?
A0A8 F0 F7		BEQ	p1
A0AA CA			DEX
A0AB D0 F8		BNE	p2

A0AD 20 16A0		JSR	StandBy

A0B0 A2 14		LDX	#20		;reasonable delay
A0B2 CA		p4	DEX
A0B3 D0 FD		BNE	p4

6502 PC/AT keyboard interface V0.1                1-Jan-87  0:01             page 4

A0B5 20 26A0		JSR	GetAT		; get the response
A0B8 C9 FE		CMP	#$FE		; compare with not ok response
A0BA F0 BA		BEQ	SendAT3		; if re-send requested  wrong go do it again
A0BC 60			RTS			;


		; send bit to keyboard, Cb is the bit to send

A0BD		SendBit
A0BD AD 60E1		LDA	ORB		; test the clock
A0C0 29 02		AND	#2
A0C2 D0 F9		BNE	SendBit		; wait for the clock to fall
A0C4 2A			ROLA			; shift Cb into A
A0C5 8D 60E1		STA	ORB		; out to to data pin

A0C8		WaitT1
A0C8 AD 60E1		LDA	ORB
A0CB 29 02		AND	#2		; test the clock line
A0CD F0 F9		BEQ	WaitT1		; wait for the clock to rise
A0CF 60			RTS

		;
		;give help reply
A0D0 C8C5CCD0		fcc	$C8,$C5,$CC,$D0
A0D4 EAEAEA		fcc	$EA,$EA,$EA
A0D7 44656D6F6E		fcc	'Demonstration program for reading PC/AT keyboard v0.1'
A0DC 7374726174
A0E1 696F6E2070
A0E6 726F677261
A0EB 6D20666F72
A0F0 2072656164
A0F5 696E672050
A0FA 432F415420
A0FF 6B6579626F
A104 6172642076
A109 302E31
A10C 0D53796E74		fcc	'\rSyntax:  PCKEY'
A111 61783A2020
A116 50434B4559
A11B 0D4F707469		fcc	'\rOptions  None',0
A120 6F6E732020
A125 4E6F6E6500
		;
		;tables
		;
A12A 08		begin	php			save flags
A12B 20 3BC0		jsr	print
A12E 0D53656E64		fcc	'\rSending keyboard reset (FF)\r',0
A133 696E67206B
A138 6579626F61
A13D 7264207265
A142 7365742028
A147 4646290D00

A14C A9 FF		lda	#$FF
A14E 20 6AA0		jsr	SendAT		; sends, then reads ACK
A151 20 3BC0		jsr	print
A154 41636B6E6F		fcc	'Acknowledgement ',0
A159 776C656467
A15E 656D656E74
A163 2000
A165 20 38C0		jsr	hexout
A168 20 2FC0		jsr	crlf

6502 PC/AT keyboard interface V0.1                1-Jan-87  0:01             page 5

A16B 20 26A0		jsr	GetAT		; An AA also expected
A16E 20 3BC0		jsr	print
A171 506F776572		fcc	'Power on Self Test Response ',0
A176 206F6E2053
A17B 656C662054
A180 6573742052
A185 6573706F6E
A18A 73652000
A18E 20 38C0		jsr	hexout
A191 20 2FC0		jsr	crlf

A194 20 3BC0		jsr	print
A197 0D53657474		fcc	'\rSetting num lock light\r',0
A19C 696E67206E
A1A1 756D206C6F
A1A6 636B206C69
A1AB 6768740D00
A1B0 A9 ED		lda	#$ED
A1B2 20 6AA0		jsr	SendAT
A1B5 20 3BC0		jsr	print
A1B8 41636B6E6F		fcc	'Acknowledgement ',0
A1BD 776C656467
A1C2 656D656E74
A1C7 2000
A1C9 20 38C0		jsr	hexout
A1CC 20 2FC0		jsr	crlf
A1CF A9 01		lda	#1
A1D1 20 6AA0		jsr	SendAT
A1D4 20 3BC0		jsr	print
A1D7 41636B6E6F		fcc	'Acknowledgement ',0
A1DC 776C656467
A1E1 656D656E74
A1E6 2000
A1E8 20 38C0		jsr	hexout
A1EB 20 2FC0		jsr	crlf

A1EE 20 3BC0		jsr	print
A1F1 5365747469		fcc	'Setting typematic (F3 53)\r',0
A1F6 6E67207479
A1FB 70656D6174
A200 6963202846
A205 3320353329
A20A 0D00
A20C A9 F3		lda	#$F3
A20E 20 6AA0		jsr	SendAT
A211 20 3BC0		jsr	print
A214 41636B6E6F		fcc	'Acknowledgement ',0
A219 776C656467
A21E 656D656E74
A223 2000
A225 20 38C0		jsr	hexout
A228 20 2FC0		jsr	crlf
A22B A9 53		lda	#$53
A22D 20 6AA0		jsr	SendAT
A230 20 3BC0		jsr	print
A233 41636B6E6F		fcc	'Acknowledgement ',0
A238 776C656467
A23D 656D656E74
A242 2000
A244 20 38C0		jsr	hexout

A247 20 3BC0		jsr	print
A24A 0D4E6F7720		fcc	'\rNow scanning\r',0
A24F 7363616E6E
6502 PC/AT keyboard interface V0.1                1-Jan-87  0:01             page 6

A254 696E670D00

A259 20 26A0	A	jsr	GetAT
A25C 20 3BC0		jsr	print
A25F 436F646520		fcc	'Code = ',0
A264 3D2000
A267 20 93A2		jsr	hx
A26A C9 F0		cmp	#$F0
A26C D0 06		bne	B
A26E 20 88A2		jsr	do
A271 4C 82A2		jmp	C

A274 C9 E0	B	cmp	#$E0
A276 D0 0A		bne	C
A278 20 88A2		jsr	do
A27B C9 F0		cmp	#$F0
A27D D0 03		bne	C
A27F 20 88A2		jsr	do

A282 20 2FC0	C	jsr	crlf
A285 4C 59A2		jmp	A

		;	plp
		;	rts

A288 A9 20	do	lda	#$20
A28A 20 23C0		jsr	out
A28D 20 26A0		jsr	GetAT
A290 4C 93A2		jmp	hx


A293 48		hx	pha
A294 20 38C0		jsr	hexout
A297 68			pla
A298 60			rts

Errors detected: 0
