{$C4500} {$X+} {$G+} {$D-}
program diel_sheet(input,output,data_file);
{This program works out transmission, reflection and absorption coefficients of a rotating dielectric in a plane wave}

label 1,10;

const pi=3.141592654;

type
      
       coeffs=array[1..6] of real;

       cursormode=record
                v          :boolean;
                mode       :integer;
                angle      :real;
                x          :coeffs;
                end;
      
      sample_info=record
                n          :real;{refractive index}
                k          :real;{dielectric loss}
                t          :real;{thickness in mm}
                w          :real;{wavelength in mm}
                end;

   var
    n1,n2,k,j                                        :integer;
    x, y                                             :coeffs;
    info                                             :sample_info;
    kap                                              :real;
    start, finish, increment, st_rad                 :real;{angle parameters}
    pts,intervals,h_pos,h_pos_last                   :integer;
    h_scale                                          :real;
    ip                                               :char;
    cmode                                            :cursormode;
    data_file                                        :text;

procedure tab(x,y:integer);
begin;
 vdu(31,x,y);
end;

procedure plot_g(newx:integer;newy:real; var oldx:integer;var 
                               oldy:real);
const vscale=924;
      v_off=50;

begin;
plot(4,oldx,trunc(v_off + oldy*vscale));
plot(5,newx,trunc(v_off + newy*vscale));
oldy:=newy;
end;

procedure clear_menu;
var y:integer;
begin;
  for y:=2 to 30 do
  begin;
  tab(71,y);write('         ');
  end;
tab(71,2);
end;

procedure main_menu(info:sample_info;pts:integer;start,finish:real);
begin;
tab(71,2);write('t=',info.t:3:1);
tab(71,4);write('n=',info.n:3:5);
tab(71,6);write('k=',info.k*(-1):3:5);
tab(71,8);write('pts=',pts:4);
tab(71,10);write('start=');
tab(71,11);write(start:3:1);
tab(71,13);write('finish=');
tab(71,14);write(finish:3:1);
tab(71,20);write('QUIT');
tab(71,22);write('PCOPY');
tab(71,24);write('CURSOR');
tab(71,26);write('SAVE');
end;

procedure pcopy;
begin;
tab(71,30);write('Pcopy Y/N');
repeat k:=inkey(50) until k>0;
if (k=ord('Y')) or (k=ord('y')) then oscli('GIMAGE EPSON');
end;


function cosh(x:real):real;
begin;
cosh:=(exp(x)+exp(-1.0*x))/2.0;
end;

function sinh(x:real):real;
begin;
sinh:=(exp(x)-exp(-1.0*x))/2.0;
end;

procedure trans_refl_coeff (theta:real;info:sample_info;
                                var x: coeffs);

var a,b,c,d,p,q,po1,po2,sin_theta,cos_theta,v0,v1,v2,v3       :real;
    n_diel_2,k_2,anga,angb                                    :real;

{routine to work out transmission and reflection coefficients for both
polarizations at an angle theta to the normal}

{equation numbers as in paper}


procedure coeff_calc(a,b,p,q,po,anga,angb :real; var Tsigma, Rsigma :real);
                             

var
va,vb,inv_pq,pa,pb,qa,qb,m1,m2,
pa_mul_qa,pb_mul_qb,Dsigma,
cosh_angb,cos_anga                                    :real; 

begin;                      

m1:=2.0*a*sinh(angb); cosh_angb:=cosh(angb);
m2:=2.0*b*sin(anga);cos_anga:=cos(anga);             

inv_pq:=1.0/(p*q);
pa:=po/(p*p)+1.0;pb:=pa-2.0;
qa:=po/(q*q)+1.0;qb:=qa-2.0;

pa_mul_qa:=pa*qa;                                              
pb_mul_qb:=pb*qb;

va:=4.0*sqr(a)*inv_pq;
vb:=4.0*sqr(b)*inv_pq;

Dsigma:=(pa_mul_qa+va)*cosh_angb + m1*(pa/q+qa/p)          
        - (pb_mul_qb-vb)*cos_anga + m2*(pb/q+qb/p);     {eqn 30}

Rsigma:=((pa_mul_qa-va)*cosh_angb + m1*(pa/q-qa/p)      {reflection coeff}
        - (pb_mul_qb+vb)*cos_anga + m2*(pb/q-qb/p))/Dsigma; {eqn 28}

Tsigma:=8.0*po*inv_pq/Dsigma;                             {transmission coeff} 
                                                        {eqn 29}

end;

{end of coeff_calc}

{start of trans_refl_coeff}

begin;

cos_theta:=cos(theta);sin_theta:=sin(theta);
p:=cos_theta;q:=p; {if refractive indices n1 and n3 are one}
n_diel_2:=sqr(info.n);k_2:=sqr(info.k);
v0:=n_diel_2 - k_2;
v1:=v0 - sqr(sin_theta);
v2:=sqrt(sqr(v1) + 4.0*n_diel_2*k_2);
a:=sqrt((v2 + v1)/2);
b:=sqrt((v2 - v1)/2);
po1:=sqr(a) + sqr(b);

anga:= 4.0*pi*a*info.t/info.w;
angb:= 4.0*pi*b*info.t/info.w;

coeff_calc(a,b,p,q,po1,anga,angb,x[4],x[5]);

v3:=sqr(n_diel_2 + k_2);
c:=(a*v0 + 2*b*info.n*info.k)/v3;
d:=(b*v0 - 2*a*info.n*info.k)/v3;
po2:=po1/v3;                                        
{p.q unchanged if n1 and n3 are one}
coeff_calc(c,d,p,q,po2,anga,angb,x[1],x[2]);
x[3]:=1 - x[1] - x[2];{a_par}
x[6]:=1 - x[4] - x[5];{a_perp}

end;

procedure save_menu(start, finish :real; info :sample_info);

type   string=packed array[1..11] of char;

var    k, f_int, z            :integer;
       d                      :coeffs;
       incr,st                :real;
       name                   :string;
       no_error               :boolean;

procedure get_name(var s:string;var b:boolean);
var    i, k                   :integer;
       f                      :boolean;
begin;
tab(71,17);write('name ?');
tab(71,18);
i:=6;
s:=':1.l.      '; {l to indicate pascal textfile of lamina data}
f:=false;
 repeat
   repeat
   k:=inkey(10);
   until ((k>27) and (k<128)) or (k=13);
  case k of
  127:if i>6 then 
      begin;
      i:=i-1;
      s[i]:=' ';
      write(chr(127):1);
      end;
  13 :f:=true;
  end {case in Acornsoft}

  otherwise if i<11 then
  begin;
  s[i]:=chr(k);
  i:=i+1;
  write(chr(k):1);
  end;

 until f;
 if i>6 then b:=true else b:=false;

end;



begin;
clear_menu;
tab(71,2);write('Save as');
tab(71,3);write('textfile');
tab(71,5);write('1.t_par');
tab(71,6);write('2-r.par');
tab(71,7);write('3-a.par');
tab(71,8);write('4-t.perp');
tab(71,9);write('5-r.perp');
tab(71,10);write('6-a.perp');
tab(71,12);write('Curve ?');
repeat k:=inkey(50) until (k>48) and (k<55);
k:=k-48;
tab(71,14);write('ivals ?');
tab(71,15);read(f_int);
get_name(name,no_error);
 if no_error then
 begin;
 rewrite(data_file,name);
 writeln(data_file,info.n);
 writeln(data_file,info.k);
 writeln(data_file,info.t);
 writeln(data_file,info.w);
 writeln(data_file,start);
 writeln(data_file,finish);
 writeln(data_file,k);{number of curve 1..6}
 writeln(data_file,f_int);{number of intervals=pts-1}

 incr:=(finish-start)*pi/(f_int*180.0);
 st:=start*pi/180.0;
  for z:=0 to f_int do
  begin;
  trans_refl_coeff(st + z*incr,info,d);
  writeln(data_file,d[k]);
  end;
 end;
end;
    

procedure cursor_menu(var c:cursormode;start,finish:real;info:sample_info);
type   pixel_colours=array[1..20] of boolean;

var    ip                        :char;
       q, r, p, onscreen         :boolean;
       k, xp, yp, z, col         :integer;
       pixel_1, pixel_2          :pixel_colours;

procedure curse(pixel:pixel_colours; xp, yp:integer);
{copies cursor to screen}
var             z:integer;
begin;
 for z:=1 to 20 do
 begin;
 plot(4,xp,yp+z);
 if pixel[z] then plot(5,xp,yp+z) else plot(7,xp,yp+z);
 end;
end;


procedure get_mode(var mo:integer);
var     k        :integer;
begin;
tab(71,29);write('mode=?');
repeat k:=inkey(10) until k>0;
if (k>48) and (k<55) then mo:=k-48 else mo:=1;
tab(71,29);write('         ');
tab(71,30);write('         ');
end; 

procedure get_angle(var ang:real);

begin;
tab(71,29);write('Angle=?');
tab(71,30);read(ang);
tab(71,29);write('       ');
tab(71,30);write('       ');
trans_refl_coeff(ang*pi/180,info,c.x);
end;


begin;
clear_menu;
tab(71,2);write('Cursor');
tab(71,4);write('1-t.par');
tab(71,5);write('2-r.par');
tab(71,6);write('3-a.par');
tab(71,7);write('4-t.perp');
tab(71,8);write('5-r.perp');
tab(71,9);write('6-a.perp');
 if not c.v then
 begin;
 c.mode:=1;
 c.angle:=(start + finish)/2.0;
 trans_refl_coeff(c.angle*pi/180,info,c.x);
 c.v:=true;
 end;
q:=false;
repeat
tab(71,16);write('mode ',c.mode:1);
tab(71,18);write('angle=');
tab(72,19);write(c.angle:5:2);
tab(71,21);write('coeff=');
tab(72,22);write(c.x[c.mode]:3:3);
tab(71,30);write('? ');
tab(72,30);
p:=true;
 repeat
   if (c.angle<=finish) and (c.angle>=start) then
   begin;
   onscreen:=true;
   xp:=50 + trunc(1050.0*c.angle/(finish-start));
   yp:=50 + trunc(c.x[c.mode]*924.0);
    for z:=1 to 20 do {put cursor on and off patterns into arrays}
    begin;
    col:=point(xp,yp+z);
    pixel_1[z]:=(col<>0);
    pixel_2[z]:=(col=0);
    end;
   end
    else onscreen:=false;
  repeat
  p:=not(p);
  k:=inkey(40);
   if onscreen then
   begin;
     if p or (k>0) then curse(pixel_1,xp,yp) else curse(pixel_2,xp,yp);
    end;
  until k>0;
  ip:=chr(k);
  r:=true;
   case ip of
   'Q','q':q:=true;
   'A','a':get_angle(c.angle);
   'M','m':get_mode(c.mode);
   end otherwise r:=false;
  until r;
until q;
end;


begin;
mode(0);
1:writeln(chr(12));
cmode.v:=false;
writeln('Program to plot transmission, reflection and absorption');
writeln('as a function of angle of incidence');
writeln('This version based on equations derived from Maxwells equations');
writeln('from Hadleigh and Dennison');
writeln;
writeln;
write('Enter wavelength <mm> ');
readln(info.w);
write('Enter thickness <mm> ');
readln(info.t);
write('Enter loss (kappa) ');
readln(kap);info.k:=kap*(-1.0);
write('Enter refractive index ');
readln(info.n);                  
write('Enter start angle <deg> ');
readln(start);
write('Enter finish angle <deg> ');
readln(finish);
write('Enter number of intervals ');
readln(intervals);
pts:=intervals + 1;
writeln(chr(12));

plot(4,50,50);
plot(1,1050,0);
plot(4,50,50);
plot(1,0,924);
 for j:=1 to 10 do
 begin;
 plot(4,50+105*j,50);
 plot(1,0,-6);
 plot(4,50,50+trunc(92.4*j));
 plot(1,-6,0);
 end;
tab(0,0);write('THEORETICAL DATA plot of coefficients against incident angle.');tab(3,31);write(start:3:1);
tab(66,31);write(finish:3:1);
tab(0,30);write('0.0');
tab(0,1);write('1.0');
main_menu(info,pts,start,finish);

increment:=(finish-start)*pi/(intervals*180);
st_rad:=start*pi/180;
h_scale:=1050.0/intervals;
trans_refl_coeff(st_rad, info, x);

h_pos:=50;
y:=x;

 for n1:=1 to intervals do
 begin;
 trans_refl_coeff(st_rad + increment*n1,info,x);
 h_pos_last:=h_pos;
 h_pos:=50 + trunc(n1*h_scale);
  for n2:=1 to 6 do
  begin;
  plot_g(h_pos,x[n2],h_pos_last,y[n2]);
  end;
 end;

10: tab(71,30);
 write('? ');
 tab(72,30);
 repeat k:=inkey(10) until k>0;
 ip:=chr(k);

  case ip of
  'Q','q':goto 1;
  'S','s':save_menu(start, finish, info);
  'C','c':cursor_menu(cmode,start,finish,info);
  'P','p':pcopy;     
  end
  
  otherwise goto 10;

clear_menu;
main_menu(info,pts,start,finish);
goto 10;
end.
 